"""
Rijnland Boezem Pump Optimization
Implemented in RTC-Tools 2.0

Authors: Ivo Miltenburg, Olav van Duin, Jesse van der Wees
"""

# Optimization modules
from rtctools.optimization.collocated_integrated_optimization_problem import CollocatedIntegratedOptimizationProblem
from rtctools.optimization.goal_programming_mixin import GoalProgrammingMixin, Goal, StateGoal
from rtctools.optimization.modelica_mixin import ModelicaMixin
from rtctools.optimization.pi_mixin import PIMixin
from rtctools.optimization.homotopy_mixin import HomotopyMixin
from rtctools.optimization.csv_lookup_table_mixin import CSVLookupTableMixin
from rtctools.optimization.timeseries import Timeseries
from rtctools.util import run_optimization_problem
from rtctools._internal.caching import cached
from rtctools._internal.alias_tools import AliasDict
import rtctools.data.pi as pi
import rtctools.data.rtc as rtc
from casadi import constpow, MX

# Simulation Modules
from rtctools.simulation.simulation_problem import SimulationProblem
from rtctools.simulation.pi_mixin import PIMixin as PIMixin_sim
from rtctools.util import run_simulation_problem

# Data Modules
from collections import OrderedDict
import numpy as np
import pandas as pd

# Useful extra funtions written by Deltares for Rijnland, kept in seperate file for readability
# Note: utilities also defines global vars BASE_FOLDER, MODEL_FOLDER, INPUT_FOLDER, OUTPUT_FOLDER, logger
from utilities import *
from emergency_closure_tools import EmergencyClosures
from polder_scheduling import schedule_polder_pumps
from pumping_analysis import execute_pumping_analysis

# Utilities
import logging
import os
import timeit
import sys
import shutil
import copy
import csv
import itertools
from multiprocessing import Process
import scipy.signal
import dill


class QP_ROC_Goal(Goal):
    def __init__(self, optimization_problem, state, priority, max_increase=np.nan, max_decrease=np.nan, sol_path_timeseries_id = None):
        self.priority = priority
        self.pump_var = state
        self.function_key = 'der({})'.format(state)

        # Set target bounds (if any)
        self.target_max = max_increase
        self.target_min = max_decrease

        if self.has_target_bounds:
            self.order = 1.0

        # Determine function_range, function_nominal
        times = optimization_problem.times()
        dt = times[1] - times[0]
        bounds = optimization_problem.bounds()[state]
        max_der = (np.max(bounds[1].values) - np.min(bounds[0].values)) / dt
        self.function_range = -max_der, max_der
        self.function_nominal = max_der / 10.0

        if sol_path_timeseries_id is not None:
            if self.has_target_bounds:
                self.violation_timeseries_id = sol_path_timeseries_id
            else:
                self.function_value_timeseries_id = sol_path_timeseries_id

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.der(self.pump_var)


class MinimizePumpCost(Goal):
    def __init__(self, pump_var, pump_sign, price_var, function_nominal, priority, relaxation=0, sol_path_timeseries_id=None):
        self.order = 1.0
        self.pump_var = pump_var
        self.pump_sign = pump_sign
        self.price_var = price_var
        self.priority = priority
        # self.function_range = 0.0, 1.0
        self.relaxation = relaxation
        self.function_nominal = function_nominal

        if sol_path_timeseries_id is not None:
            if self.has_target_bounds:
                self.violation_timeseries_id = sol_path_timeseries_id
            else:
                self.function_value_timeseries_id = sol_path_timeseries_id

    def function(self, optimization_problem, ensemble_member):
        op = optimization_problem
        # get collocation times of self.pump_var
        times = op.times(self.pump_var)

        dt_array = np.diff(times)
        total_time = op.times()[-1] - op.times()[0]

        total_cost = 0.0
        for t, dt in zip(times[1:], dt_array):
            total_cost += self.pump_sign * (dt / total_time) * op.state_at(self.pump_var, t) * op.state_at(self.price_var, t)
        return total_cost


class WaterLevelRangeGoal(Goal):
    def __init__(self, optimization_problem, state, state_bounds, target_min, target_max, priority, relaxation=0.0, critical=False, weight=1.0, sol_path_timeseries_id=None):
        self.state = state
        self.weight = weight
        self.target_min = target_min
        self.target_max = target_max
        self.priority = priority
        self.relaxation = relaxation
        self.critical = critical
        self.function_range = state_bounds
        self.function_nominal = abs((state_bounds[1] + state_bounds[0]) / 2.0)
        self.function_key = optimization_problem.signed_canonical_state(state)

        if sol_path_timeseries_id is not None:
            if self.has_target_bounds:
                self.violation_timeseries_id = sol_path_timeseries_id
            else:
                self.function_value_timeseries_id = sol_path_timeseries_id

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)


class WaterLevelSetpointGoal(Goal):
    def __init__(self, optimization_problem, state, state_bounds, setpoint, setpoint_band, priority, relaxation=0, critical=False, weight=1.0, sol_path_timeseries_id = None):
        self.state = state
        self.weight = weight
        self.target_min = Timeseries(setpoint.times, setpoint.values - setpoint_band)
        self.target_max = Timeseries(setpoint.times, setpoint.values + setpoint_band)
        self.priority = priority
        self.relaxation = relaxation
        self.critical = critical
        self.function_range = state_bounds
        self.function_nominal = abs((state_bounds[1] + state_bounds[0]) / 2.0)
        self.function_key = optimization_problem.signed_canonical_state(state)

        if sol_path_timeseries_id is not None:
            if self.has_target_bounds:
                self.violation_timeseries_id = sol_path_timeseries_id
            else:
                self.function_value_timeseries_id = sol_path_timeseries_id

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)


class ChlorideRangeGoal(Goal):
    order = 1.0
    def __init__(self, optimization_problem, state, target_max, priority, sol_path_timeseries_id = None, relaxation=0):
        self.state = state
        self.target_max = target_max
        self.priority = priority
        max_c = np.nanmax(target_max.values)
        self.function_range = 0.0, 1.0
        self.function_nominal = max_c
        self.relaxation = relaxation
        self.function_key = optimization_problem.signed_canonical_state(state)

        if sol_path_timeseries_id is not None:
            if self.has_target_bounds:
                self.violation_timeseries_id = sol_path_timeseries_id
            else:
                self.function_value_timeseries_id = sol_path_timeseries_id

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.state)


class DischargeGoal(Goal):
    def __init__(self, optimization_problem, state, function_range, priority, target_min=np.nan, target_max=np.nan,
                 order=1, target_relaxation=0.0, sol_path_timeseries_id=None):
        self.pump_var = state

        # Apply target_relaxation and set targets
        if isinstance(target_min, Timeseries):
            self.target_min = Timeseries(target_min.times, target_min.values[:] - target_relaxation)
        else:
            self.target_min = target_min - target_relaxation
        if isinstance(target_max, Timeseries):
            self.target_max = Timeseries(target_max.times, target_max.values[:] + target_relaxation)
        else:
            self.target_max = target_max + target_relaxation

        self.function_range = function_range
        self.function_nominal = np.max(np.abs(function_range)) / 10.0
        self.priority = priority
        self.function_key = optimization_problem.signed_canonical_state(state)
        self.order = order

        if sol_path_timeseries_id is not None:
            if self.has_target_bounds:
                self.violation_timeseries_id = sol_path_timeseries_id
            else:
                self.function_value_timeseries_id = sol_path_timeseries_id

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state(self.pump_var)


class MinimizeQpumpGoal(Goal):
    def __init__(self, optimization_problem, pump_var, function_range, priority=2, relaxation=0, sol_path_timeseries_id = None):
        self.pump_var = pump_var
        self.function_range = function_range
        # TODO: Fix double scaling?
        self.function_nominal = np.max(np.abs(function_range)) / 10
        self.priority = priority
        self.relaxation = relaxation

        if sol_path_timeseries_id is not None:
            if self.has_target_bounds:
                self.violation_timeseries_id = sol_path_timeseries_id
            else:
                self.function_value_timeseries_id = sol_path_timeseries_id

    def function(self, optimization_problem, ensemble_member):
        # The penalty variable is taken to the 2nd power so negative values
        # are also minimized to close to 0.
        return optimization_problem.state(self.pump_var) ** 2

    order = 1


class Rijnland_base:
    # Use this class to store some basic properties that can be reused in
    # different run modes.

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super(Rijnland_base, self).__init__(**kwargs)

        # Construct list of active reaches
        try:
            with open(os.path.join(kwargs['input_folder'], 'closed_reaches.csv'), 'r') as csv_file:
                reader = csv.reader(csv_file)
                closed_reaches = list(reader)[0]
            self.closed_reaches = closed_reaches
            self.active_reaches = [r for r in self.reaches if r not in closed_reaches]
        except IOError:
            self.active_reaches = self.reaches
            self.closed_reaches = []

        # Populate list of additional outputs
        self.additional_output = []
        for reach in self.active_reaches:
            self.additional_output.append(reach + '.HQUp.H')
            self.additional_output.append(reach + '.HQDown.H')
            self.additional_output.append(reach + '.HQUp.Q')
            self.additional_output.append(reach + '.HQDown.Q')
            self.additional_output.append(reach + '.Q[2]')

        for polder in self.polders:
            self.additional_output.append('PG_{}_Q'.format(polder))

        for storage_node in self.storage_nodes:
            self.additional_output.append(storage_node + '.HQ.H')
            # self.additional_output.append(storage_node + '.HQ.Q')
            self.additional_output.append(storage_node + '.HQ.C')

    pi_parameter_config_basenames = [
        'rtcParameterConfig_Boezem',
        'rtcParameterConfig_Boezemgemalen',
        'rtcParameterConfig_Kanalen',
        'rtcParameterConfig_Goals',
        'rtcParameterConfig_Polders',
        'rtcParameterConfig_Knopen',
        'rtcParameterConfig_Boezem-PID']

    reaches = [
        'Aarkanaal_1', 'Bennebroekervaart_1', 'Braassemermeer_1', 'Braassemermeer_2',
        'Buiten_Liede_1', 'Does_1', 'Drecht_1', 'Drecht_2', 'Fuikvaart_1', 'Gouwe_2',
        'Gouwe_3', 'Gouwe_4', 'Halfweg_2', 'Leidsevaart_1', 'Leidsevaart_2', 'Leidsevaart_3',
        'Nieuwe__Rijn_1', 'Nieuwe_Vaart_1', 'Oegstgeesterkanaal_2',
        'Oegstgeesterkanaal_3', 'Oude_Rijn_1', 'Oude_Rijn_2', 'Oude_Rijn_3', 'Oude_Rijn_4',
        'Oude_Wetering_1', 'Rijn_1', 'Rijn_Schiet_1', 'Rijn_schiet_2', 'Rijn_Schiet_3',
        'Ringvaart_1', 'Ringvaart_2', 'Ringvaart_3', 'Ringvaart_4', 'Ringvaart_5',
        'Ringvaart_6', 'Ringvaart_7', 'Ringvaart_8', 'Ringvaart_9', 'Spaarne_2', 'Spaarne_3',
        'Spaarne_4', 'Spaarne_5', 'Vliet_1', 'Vliet_2', 'Vliet_3', 'Warmonderleede_1',
        'Weipoortse_Vliet_1', 'Woudwetering_1', 'Nieuwe_Meer_1']

    storage_nodes = [
        'BG_Gouda', 'BG_Halfweg', 'BG_Katwijk', 'BG_Spaarndam', 'ML_Amsterdam_NM',
        'ML_Alphen','ML_Boskoop', 'ML_Heemstede_A', 'ML_Heemstede_C',
        'ML_Leiden', 'ML_Leidschendam','ML_Lisse', 'ML_Nieuwe_Wetering', 'ML_Sloten',
        'ML_Tolhuissluis', 'SN_Bodegraven', 'SN_01', 'SN_02', 'SN_03', 'SN_04',
        'SN_05', 'SN_06', 'SN_07', 'SN_08', 'SN_09', 'SN_10', 'SN_11', 'SN_12',
        'SN_13', 'SN_14', 'SN_15', 'SN_16', 'SN_17', 'SN_18', 'SN_19', 'SN_20',
        'SN_21', 'SN_22', 'Eindk_Haarlemmermeer', 'Eindk_N3MP']

    measured_nodes = [
        'BG_Gouda', 'BG_Halfweg', 'BG_Katwijk', 'BG_Spaarndam', 'ML_Amsterdam_NM',
        'ML_Alphen','ML_Boskoop', 'ML_Heemstede_A', 'ML_Heemstede_C',
        'ML_Leiden', 'ML_Leidschendam','ML_Lisse', 'ML_Nieuwe_Wetering', 'ML_Sloten',
        'ML_Tolhuissluis', 'Eindk_Haarlemmermeer', 'Eindk_N3MP']

    # A list of all the names of the polders
    polders = [
        'afw_246',
        'afw_249']

    # TODO: Read from CSV
    pump_stages = {
        'GEM_Spaarndam': np.array([[0, -0.0], [1, -16.0], [2, -32.0], [3, -34.0]]),
        'GEM_Halfweg': np.array([[0, -0.0], [1, -7.3], [4, -22.0], [8, -33.0]]),
        'GEM_Gouda': np.array([[0, -0.0], [1, -8.33], [2, -13.0], [3, -16.0], [4, -21.0], [5, -26.0], [6, -34.0], [7, -40.0]]),
        'GEM_Gouda_inflow': np.array([[0, 0.0], [8, 11.0], [9, 25.0], [10, 35.0]]),
        'GEM_Katwijk': np.array([[0, -0.0], [1, -15.0], [2, -30.0], [3, -45.0], [4, -60.0], [5, -75.0], [6, -94.0]]),
        'GEM_N3MP': np.array([[0, -0.0], [1, -10.0], [2, -20.0], [3, -30.0], [4, -40.0], [5, -50.0], [6, 10.0], [7, 20.0], [8, 30.0], [9, 40.0], [10, 50.0]]),
        'GEM_Haarlemmermeer': np.array([[0, -0.0], [1, -10.0], [2, -20.0], [3, -30.0], [4, -40.0], [5, -50.0], [6, 10.0], [7, 20.0], [8, 30.0], [9, 40.0], [10, 50.0]])
    }

    pump_sign = {pump: np.sign(stages[-1, 1]) for pump, stages in pump_stages.items()}

    # Define pump categories
    emergency_pumps = [
        'GEM_Haarlemmermeer',
        'GEM_N3MP']
    boezem_pumps = [
        'GEM_Spaarndam',
        'GEM_Katwijk',
        'GEM_Halfweg',
        'GEM_Gouda',
        'GEM_Gouda_inflow']
    pumps = boezem_pumps + emergency_pumps


class Rijnland_Optimization(Rijnland_base, HomotopyMixin, GoalProgrammingMixin, PIMixin, ModelicaMixin,
               CollocatedIntegratedOptimizationProblem):

    model_name = 'NetworkOptimization'

    # Spaarndam and Halfweg have the same priority and can be used interchangebly
    pump_priorities = {
        'GEM_Gouda': 50,
        'GEM_Gouda_inflow': 50,
        'GEM_Katwijk': 51,
        'GEM_Spaarndam': 52,
        'GEM_Halfweg': 52}

    # Fill be filled during runtime
    pump_priorities_active = set()

    # We save these goals separately to be able to do the pumping analysis
    pump_cost_goals       = []
    pump_cost_path_goals  = []

    #TODO: get this from parameters?
    _H_state_bounds = (-1.5, -0.2)

    # Prepare dicts for additional variables, will be filled later
    _UnitPrice_vars = {}

    # Prepare dicts for additional bounds, will be filled later
    _Q_pump_bounds = {}
    _UnitPrice_bounds = {}
    _H_reach_bounds = {}

    # FIXME: this is needed because otherwise an error is thrown when the bounds
    # are read during the pre method.
    _HomotopyMixin__theta = 0

    # Use a small tolerance for checking water level range deviations
    level_tol = 1e-3 # 1 mm

    # Define which bounds to use
    use_Q_pump_bounds_from_timeseries = True

    # Dict to store sol path info
    sol_path_dict = {}

    # Output sol_path.csv after each priority finishes (or just wait until the end)
    output_sol_path_after_each_prio = True

    use_polder_level_goals = True

    # access homotopy parameter using self.h_th
    @property
    def h_th(self):
        return self.parameters(0)['theta']

    def signed_canonical_state(self, state):
        canonical, sign = self.alias_relation.canonical_signed(state)
        return canonical if sign > 0.0 else '-' + canonical

    def pre(self):
        # use timeseries_import with initial states calculated
        self.timeseries_import_basename = 'timeseries_import-initialized'

        super().pre()

        self._post_first_priority = False

        parameters = self.parameters(0)

        self.streefpeil_RBP = parameters['RBP_H_STREEF']
        self.ondergrens_RBP = parameters['RBP_H_ONDER']
        self.bovengrens_RBP = parameters['RBP_H_BOVEN']
        # We actually allow a bit more variation around the setpoint.
        self.streefpeil_RBP_band = 0.001

        # Tolarance for checking flow rate compliance
        self.q_req_tol = parameters.get('q_req_tol', 0.5)

        # A little bit of tolerance when checking for acceptance, because
        # strictly speaking -0.4999... is smaller than 0.5.
        self._H_min = self.ondergrens_RBP - self.level_tol
        self._H_max = self.bovengrens_RBP + self.level_tol
        self._H_sp_min = self.streefpeil_RBP - self.streefpeil_RBP_band - self.level_tol
        self._H_sp_max = self.streefpeil_RBP + self.streefpeil_RBP_band + self.level_tol

        # Get goal parameters - which goals are available to be used in the
        # different scenarios
        self.use_H_range_goal = parameters['use_H_range_goal']
        self.use_Hnode_range_goal = parameters['use_Hnode_range_goal']
        self.use_H_setpoint_goal = parameters['use_H_setpoint_goal']
        self.use_Q_emergency_pump_minimize_goal = parameters['use_Q_emergency_pump_minimize_goal']
        self.use_pump_Hdown_range_goal = parameters['use_pump_Hdown_range_goal']
        self.use_Cl_range_goal = parameters['use_Cl_range_goal']
        self.use_Cl_flushing_goal = parameters['use_Cl_flushing_goal']
        self.use_Q_pump_req_goal = parameters['use_Q_pump_req_goal']
        self.use_katwijk_tide_goals = parameters['use_katwijk_tide_goals']
        self.use_cost_pump_minimize_goal_var = parameters['use_cost_pump_minimize_goal_var']
        self.prioritize_between_pumps = parameters['prioritize_between_pumps']

        # Switch between mode where minimization of pump use has a different
        # priority per pump, or the same priority for all.
        if not self.prioritize_between_pumps:
            # Find the lowest priority of the original settings
            prio = min([i for i in self.pump_priorities.values()])
            # Set all priorities to that value.
            for pump in self.boezem_pumps:
                self.pump_priorities[pump] = prio

        # Set the pumping analysis. Option are 'subprocess', 'serial' or 'none'
        self.pumping_analysis = parameters['pumping_analysis']

        for pump in self.boezem_pumps:
            logger.info("{}: priority = {}".format(pump, str(self.pump_priorities[pump])))

        # Set up inputs for GEM_Gouda_inflow (this is because FEWS only knows about one pump at Gouda)
        pump, pump_inflow = 'GEM_Gouda', 'GEM_Gouda_inflow'
        copied_timeseries = ['_Qmax', '_Qmin', '_Q0_guess']
        for ts in copied_timeseries:
            try:
                self.set_timeseries(pump_inflow + ts, self.get_timeseries(pump + ts))
            except KeyError:
                continue

        # Inflow pump costs zero at all timesteps
        self.set_timeseries(pump_inflow + '_UnitPrice_m3s', Timeseries(self.times(), np.zeros_like(self.times())))

        #  Get delta theta
        # TODO: This variable is not used because homotopy_options() calls parameters and there are recursion issues
        self.delta_theta_0 = parameters['delta_theta_0']

        # Timestep settings
        self.adjust_timesteps = parameters.get('adjust_timesteps', False)
        self.debug_timestep_rescaling = parameters.get('debug_timestep_rescaling', False)
        self.equidistant_export_times = parameters.get('equidistant_export_times', False)

        ########## Adjust Timesteps #########

        if self.adjust_timesteps:
            logger.info('Adjusting timeseries timesteps, be careful when interpreting the results!')
            # TODO: automatically scale ranges and timesteps?
            logger.warning(
                'Setting adjust_timesteps to True requires adjusting function ranges and nominals of path goals')

            # Store timeseries times for use during post() export
            # (to avoid having to reconfigure FEWS to accept non-equidistant timeseries)
            self.equidistant_timeseries_import = copy.deepcopy(self.timeseries_import)
            self.equidistant_datetimes = self.timeseries_import.times[self.timeseries_import.forecast_index:]
            self.equidistant_times = self._datetime_to_sec(self.timeseries_import.times)
            self.equidistant_deltatime = self.timeseries_import.dt

            equidistant_times = self.equidistant_times
            equidistant_timestep = self.timeseries_import.dt.total_seconds()

            # We want to keep all of the first 12 steps, 4 of the next 12 steps, and 2 of the
            # subsequent sets of 12 steps. For a 1 hr timestep, the timesteps are reduced to
            # every 3 hours and every 6 hours thereafter

            def reduced_timesteps(series, chunk_len = 12):
                # series len must be a multiple of chunk_len
                try:
                    assert(len(series) % chunk_len == 0)
                except AssertionError:
                    logger.error("Optimization period must contain a multiple of {} timesteps (including T0!).".format(chunk_len))
                    raise
                # chunk_len must be divisible by 3
                try:
                    assert(chunk_len % 3 == 0)
                except AssertionError:
                    logger.error("Chunk length in reduce timesteps mode must be a multiple of 3).")
                    raise
                chunks = np.reshape(series, (-1, chunk_len))
                adjusted_chunks = []
                for i, c in enumerate(chunks):
                    if i == 0:
                        adjusted_chunks.extend(c)
                    elif i == 1:
                        adjusted_chunks.extend(c.reshape((-1,3)).mean(-1))
                    else:
                        adjusted_chunks.extend(c.reshape((-1,6)).mean(-1))
                return adjusted_chunks

            # Overwrite timeseries_import times with adjusted times
            self.timeseries_import._times = self._sec_to_datetime(reduced_timesteps(equidistant_times))

            # Overwrite timeseries_import values with interpolated values
            for var_name, values in self.timeseries_import.items():
                values = reduced_timesteps(values)
                self.timeseries_import.set(var_name, values)

            # Write changes to time-related variables
            self.timeseries_import.dt = None
            self.timeseries_import_times = self._datetime_to_sec(self.timeseries_import.times)

            logger.info('TimestepRescalingMixin: Done adjusting timeseries timesteps.')

            self.timeseries_import_dict = [AliasDict(self.alias_relation) for ensemble_member in range(self.ensemble_size)]
            for ensemble_member in range(self.ensemble_size):
                for key, value in self.timeseries_import.items(ensemble_member):
                    self.timeseries_import_dict[ensemble_member][key] = value

            # These new timestps can be written to timeseries import if debugging/validation is needed
            if self.debug_timestep_rescaling:
                logger.info('TimestepRescalingMixin: Overwriting Input timeseries.')
                self.timeseries_import.write()

        ########## Done Adjusting Timesteps #########


        # Get bounds
        bounds = self.bounds()

        # Set bounds on reaches
        for reach in self.active_reaches:
            self._H_reach_bounds[reach + '.HQUp.H'] = self._H_state_bounds
            self._H_reach_bounds[reach + '.HQDown.H'] = self._H_state_bounds
        self._H_reach_bounds['RBP_H'] = self._H_state_bounds

        # Overrule bounds in mo file with bounds from FEWS if available
        if self.use_Q_pump_bounds_from_timeseries:
            # Get basename_import from PI mixin for logging
            logger.info("Reading bounds on pump discharges from {}".format(self.timeseries_import_basename))

            for pump in self.pumps:
                # The highest possible pump flow is stored as a negative number in
                # FEWS, but called 'max'. Here we have to use that as the lower bound.
                m = self.get_timeseries(pump + '_Qmax')

                # The lowest possible pump flow is stored as a positive number in
                # FEWS, but called 'min'. Here we have to use that as the upper bound.
                M = self.get_timeseries(pump + '_Qmin')

                try:
                    # First check if param exists
                    param_min = parameters[pump + '_Q_MAX'] # Q_MAX is actually a min
                except KeyError:
                    pass
                else:
                    # Adjust bounds
                    m.values[m.values < param_min] = param_min
                    self.set_timeseries(pump + '_Qmax', m)

                stages = self.pump_stages.get(pump, None)
                if stages is not None:
                    stages = stages[:,1]
                    m = snap_to_discrete(m, stages, strategy='up')
                    self.set_timeseries(pump + '_Qmax', m)
                    M = snap_to_discrete(M, stages, strategy='down')
                    self.set_timeseries(pump + '_Qmin', M)

                # Save the new bounds in a dictionary. Will be added to the global
                # bounds when needed.
                # TODO: make sure that equal or very close upper and lower bounds do not cause problems
                if pump in self.emergency_pumps and self.use_Q_emergency_pump_minimize_goal == False:
                    # Set bounds to zero, because if use_Q_emergency_pump_minimize_goal == False,
                    # these pumps are not intended to be used.
                    self._Q_pump_bounds[pump + '_Q'] = (0, 0)
                else:
                    # Store bounds in dict.
                    self._Q_pump_bounds[pump + '_Q'] = (m, M)
                #TODO: investigate why the timeseries of GEM_Haarlemmermeer_Qoutput
                # and GEM_N3MP_Qoutput are not available for the PIMixin when
                # they are not set to zero here.
                # Check if emergency pumps are bounded to (0,0)
                if pump in self.emergency_pumps and not(m.values + M.values).any():
                    dummy_series = Timeseries(m.times, np.full_like(m.values, 0))
                    self.set_timeseries(pump + '_Qoutput', dummy_series)


        # Gouda is not allowed to pump water into the system when the
        # chloride concentration in the Hollandse IJssel exceeds a
        # certain level. Overrule previously set bounds here.

        # Read (maximum) chloride levels
        Cl_max = self.get_timeseries('Eindk_Gouda_Cmax_buiten').values
        Cl = self.get_timeseries('Eindk_Gouda_C').values

        # Determine current upper bound
        pump_var = 'GEM_Gouda_inflow_Q'
        if self.use_Q_pump_bounds_from_timeseries:
            q_min = self._Q_pump_bounds[pump_var][0].values
            q_max = self._Q_pump_bounds[pump_var][1].values
        else:
            q_min = np.full_like(self.times(), bounds[pump_var][0])
            q_max = np.full_like(self.times(), bounds[pump_var][1])

        # No inflow allowed when Cl > Cl_max
        # TODO: make sure that equal or very close upper and lower bounds do not cause problems
        mask = Cl > Cl_max
        q_min[mask] = 0.0
        q_max[mask] = 0.0

        # Construct new bound timeseries and save in _Q_pump_bounds. Will be added to the global
        # bounds when needed.
        self._Q_pump_bounds[pump_var] = (
            Timeseries(self.times(), q_min),
            Timeseries(self.times(), q_max))

        # Add price signal per pump as variables to list if available, and
        # determine bounds.
        if self.use_cost_pump_minimize_goal_var:
            logger.info("Adding unit price variables for the minimize pump cost goal.")
        for pump in self.pumps:
            # Construct price variable name
            price_var = pump + '_UnitPrice_m3s'

            try:
                price = self.get_timeseries(price_var)
                # Save the variables in a dictionary. We add the variable
                # to the model in the path_variables() method
                self._UnitPrice_vars[price_var] = MX.sym(price_var)
            except KeyError:
                # If either the timeseries or the symbol is not found, continue
                continue
            else:
                # We make the bounds equal to enforce an equality
                # constraint. We add the bounds to the global bounds in the bounds() method
                self._UnitPrice_bounds[price_var] = (price, price)

        if self.use_katwijk_tide_goals:
            # Set up a request of 0.0 for GEM_Katwijk when tide is rising to squish tide towards the falling edge
            # WARNING: This is a hacky and indirect way of getting the pump to pump at times of reduced head.
            #          The correct way of doing this would be to make sure the energy price timeseries is reflects
            #          the reduced pumping costs of lower tides. If this is not the case yet, it really should be.
            last_index = len(self.times()) - 1
            try:
                [peak_tide_indices] = np.where(np.isfinite(self.get_timeseries('Eindk_Katwijk_Hpeak').values))
                [low_tide_indices] = np.where(np.isfinite(self.get_timeseries('Eindk_Katwijk_Hlow').values))
            except KeyError:
                logger.warning('Katwijk Tide goal KeyError: Eindk_Katwijk_Hpeak, Eindk_Katwijk_Hlow timeseries not complete.')
                tide = self.get_timeseries('Eindk_Katwijk_H').values
                # If the timeseries are missing, guess
                [peak_tide_indices] = scipy.signal.argrelmax(tide, order=4)
                [low_tide_indices] = scipy.signal.argrelmin(tide, order=4)

            try:
                # Find inices where katwijk tide is rising
                if low_tide_indices[0] != 0 and peak_tide_indices[0] != 0:
                    # first step is unknown- need to decide what the first step is
                    if low_tide_indices[0] < peak_tide_indices[0]:
                        # for the purposes of this run, treat first step like a high tide
                        peak_tide_indices = np.insert(peak_tide_indices, 0, 0)
                    else:
                        # for the purposes of this run, treat first step like a low tide
                        low_tide_indices = np.insert(low_tide_indices, 0, 0)

                if low_tide_indices[-1] != last_index and peak_tide_indices[-1] != last_index:
                    # last step is unknown- need to decide what the last step is
                    if low_tide_indices[-1] > peak_tide_indices[-1]:
                        # for the purposes of this run, treat last step like a high tide
                        peak_tide_indices = np.append(peak_tide_indices, last_index)
                    else:
                        # for the purposes of this run, treat last step like a low tide
                        low_tide_indices = np.append(low_tide_indices, last_index)

                # Find times between low and high tides
                if low_tide_indices[0] == 0:
                    rising_tide_indices_dummy = np.concatenate(
                        [range(l, h + 1) for l, h in zip(low_tide_indices, peak_tide_indices)])
                    # rising_tide_indices_derQmax = np.concatenate(
                    #     [range(l, h) for l, h in zip(low_tide_indices, peak_tide_indices)])
                else:
                    rising_tide_indices_dummy = np.concatenate(
                        [range(l, h + 1) for l, h in zip(low_tide_indices, peak_tide_indices[1:])])
                    # rising_tide_indices_derQmax = np.concatenate(
                    #     [range(l, h) for l, h in zip(low_tide_indices, peak_tide_indices[1:])])
                # convert all numbers in rising_tides array to integers, as the will be used as index
                rising_tide_indices = [int(i) for i in rising_tide_indices_dummy]

            except IndexError:
                logger.error('Katwijk Tide goal IndexError: Eindk_Katwijk_Hpeak, Eindk_Katwijk_Hlow: timeseries not complete.')
                raise

            # Make the req ts
            try:
                q_req = np.full_like(self.times(), np.nan)
                q_req[rising_tide_indices] = 0.0
                self.set_timeseries('GEM_Katwijk_Qreq_tide', q_req)
            except IndexError:
                logger.error('Katwijk Tide Qreq IndexError: Eindk_Katwijk_Hpeak, Eindk_Katwijk_Hlow: timeseries not complete.')
                raise

            # Create a GEM_Katwijk_derQmax_tide time series with 0's where eindk_Katwijk H (tide) is 'threshold' above the minimum tides
            der_katwijk_max = np.full_like(self.times(), np.nan)
            threshold = parameters.get('GEM_Katwijk_threshold', None)
            if threshold is not None:
                threshold_ts = self.get_timeseries('Eindk_Katwijk_Hlow').values + threshold
                nans, x = np.isnan(threshold_ts), lambda z: z.nonzero()[0]
                threshold_ts[nans]= np.interp(x(nans), x(~nans), threshold_ts[~nans])
                # logger.info(threshold_ts)
                tide_ts = self.get_timeseries('Eindk_Katwijk_H')
                der_katwijk_max[tide_ts.values > threshold_ts] = 0.0
                # now set the der_katwijk_max values at the rising edge back to nan
                # der_katwijk_max[rising_tide_indices_derQmax] = np.nan
            # Note: Pump direction is negative- modify carefully
            self.set_timeseries('GEM_Katwijk_derQmax_tide', der_katwijk_max)
            # logger.info(der_katwijk_max)

        # Set Polder Inflows If Missing (TODO: Remove)
        for polder in self.polders:
            try:
                self.get_timeseries('PG_{}_Q'.format(polder))
            except KeyError:
                self.set_timeseries('PG_{}_Q'.format(polder), np.full_like(self.times(), 0.0))
            try:
                self.get_timeseries('BUF_{}_C'.format(polder))
            except KeyError:
                self.set_timeseries('BUF_{}_C'.format(polder), np.full_like(self.times(), 0.0))

    @cached
    def linearization_parameters(self):
        # Get results of initial simulation

        presim_results = globals().get('FULL_PRESIM_RESULTS', getattr(self, '_full_presim_results', None))

        assert presim_results is not None, 'FULL_PRESIM_RESULTS could not be found'

        # Use results of initial simulatio to set linearization parameters
        linearization_parameters = AliasDict(self.alias_relation)
        for reach in self.active_reaches:
            logger.debug('Linearization Parameters:')
            q_mean = np.mean(presim_results.get(reach + '.Q[2]'))
            linearization_parameters[reach + '_Q_nominal'] = -1.0 if q_mean < 0.0 else 1.0
            logger.debug('{}_Q_nominal: {}'.format(reach, linearization_parameters[reach + '_Q_nominal']))
            c_up = np.mean(presim_results.get(reach + '.HQUp.C'))
            c_down = np.mean(presim_results.get(reach + '.HQDown.C'))
            linearization_parameters[reach + '_C_nominal'] = np.mean(np.append(c_up, c_down))
            logger.debug('{}_C_nominal: {}'.format(reach, linearization_parameters[reach + '_C_nominal']))
            h_up = np.mean(presim_results.get(reach + '.HQUp.H'))
            h_down = np.mean(presim_results.get(reach + '.HQDown.H'))
            linearization_parameters[reach + '_H_nominal'] = np.mean(np.append(h_up, h_down))
            logger.debug('{}_H_nominal: {}'.format(reach, linearization_parameters[reach + '_H_nominal']))

        for reach in self.closed_reaches:
            linearization_parameters[reach + '_Q_nominal'] = 0.0
            linearization_parameters[reach + '_C_nominal'] = 0.0
            linearization_parameters[reach + '_H_nominal'] = 0.0

        for pump in self.pumps:
            if self.pump_sign.get(pump, -1.0) < 0.0:
                linearization_parameters[pump + '_C_nominal'] = np.mean(presim_results.get(pump + '.HQDown.C'))
            else:
                linearization_parameters[pump + '_C_nominal'] = np.mean(presim_results.get(pump + '.HQUp.C'))
            logger.debug('{}_C_nominal: {}'.format(pump, linearization_parameters[pump + '_C_nominal']))

        for sn in self.storage_nodes:
            linearization_parameters[sn + '_V_nominal'] = np.mean(presim_results.get(sn + '.V'))
            logger.debug('{}_V_nominal: {}'.format(sn, linearization_parameters[sn + '_V_nominal']))
            linearization_parameters[sn + '_C_nominal'] = np.mean(presim_results.get(sn + '.HQ.C'))
            logger.debug('{}_C_nominal: {}'.format(sn, linearization_parameters[sn + '_C_nominal']))

        return linearization_parameters

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)
        parameters.update(self.linearization_parameters())
        return parameters

    @property
    def path_variables(self):
        variables = super().path_variables[:]
        # Extend with unit price bounds
        for key, variable in sorted(self._UnitPrice_vars.items()):
            variables.append(variable)
        return variables

    def bounds(self):
        bounds = super().bounds()
        # TODO? make sure that we merge, not overwrite, bounds of of aliased states
        # times = self.times(key)
        # if lb is not None:
        #     if isinstance(lb, Timeseries):
        #         lb = self.interpolate(times, lb.times, lb.values, -np.inf, -np.inf)
        #     elif isinstance(lb, ca.MX):
        #         lb = float(lb)
        # if ub is not None:
        #     if isinstance(ub, Timeseries):
        #         ub = self.interpolate(times, ub.times, ub.values, +np.inf, +np.inf)
        #     elif isinstance(ub, ca.MX):
        #         ub = float(ub)

        # # Check if we are overwriting bounds of an aliased state
        # aliased_bounds = resolved_bounds.get(key, None)
        # if aliased_bounds is not None:
        #     # The intersection of all bound ranges applies
        #     if lb is not None:
        #         lb = ca.fmax(lb, -np.inf if aliased_bounds[0] is None else aliased_bounds[0])
        #     else:
        #         lb = aliased_bounds[0]
        #     if ub is not None:
        #         ub = ca.fmin(ub, np.inf if aliased_bounds[1] is None else aliased_bounds[1])
        #     else:
        #         ub = aliased_bounds[1]
        bounds.update(self._Q_pump_bounds)
        bounds.update(self._UnitPrice_bounds)
        bounds.update(self._H_reach_bounds)
        return bounds

    # Potentially beneficial formulation improvement:
    # def path_objective(self, ensemble_member):
    #     path_objective = super(Rijnland_Homotopy_Optimization, self).path_objective(ensemble_member)
    #     regularization = 1e-6
    #     for state in itertools.chain(self.dae_variables['states'], self.dae_variables['algebraics'], self.dae_variables['control_inputs']):
    #         path_objective += regularization * constpow(state / self.variable_nominal(state.getName()), 2)
    #     return path_objective

    def constraints(self, ensemble_member):
        c = super().constraints(ensemble_member)
        # Constrain the initial pump flow rate. If not provided, then zero
        # Note: shouldn't influence the model when the discretization is backwards Euler
        for pump in self.pumps:
                try:
                    q0 = self.get_timeseries(pump + '_Q0_guess').values[0]
                    # Ensure initial condition is within bounds
                    m, M = self._Q_pump_bounds[pump + '_Q']
                    if type(m) is Timeseries:
                        m = m.values[0]
                    if type(M) is Timeseries:
                        M = M.values[0]
                    q0 = float(q0 if q0 < M else M)
                    q0 = float(q0 if q0 > m else m)
                except KeyError:
                    # If there are any KeyErrors, we constrain to a safe value of 0.0
                    q0 = 0.0
                c.append((self.state_at(pump + '_Q', 0.0) , q0, q0))
        for reach in self.active_reaches:
            c.append((self.state_at(reach + '.Q[1]', 0.0) - self.state_at(reach + '.Q[2]', 0.0), 0.0, 0.0))
            c.append((self.state_at(reach + '.Q[2]', 0.0) - self.state_at(reach + '.Q[3]', 0.0), 0.0, 0.0))
        return c

    @cached
    def bounds_ok(self):
        # Do check on bounds of all pumps.
        # TODO: The logic in this method could be a lot cleaner/simpler

        # Read bounds
        bounds = self.bounds()

        bounds_ok = {}
        for pump in self.pumps:
            b = bounds[pump + '_Q']
            min_max_bound_buffer = 1e-3
            if isinstance(b[0], Timeseries):
                # Check for missings in bounds.
                if np.isnan(np.sum(b[0].values)) or np.isnan(np.sum(b[1].values)):
                    logger.error("Bounds on discharge of {} contain at least one missing value. This is not allowed".format(pump))
                    raise Exception

                # Check if bounds are wide enough
                if not np.allclose(b[0].values, b[1].values, atol=min_max_bound_buffer, rtol=0.0):
                    bounds_ok[pump] = True
                else:
                    bounds_ok[pump] = False
                    logger.info("Bounds on discharge of {} are too close together for optimization. No goals will be added for this pump.".format(pump))
            else:
                # Check for missings in bounds.
                if np.isnan(b[0]) or np.isnan(b[1]):
                    logger.error("Bounds on discharge of {} contain at least one missing value. This is not allowed".format(pump))
                    raise Exception

                # Check if bounds are wide enough
                if not np.isclose(1.0*b[0], 1.0*b[1], atol=min_max_bound_buffer, rtol=0.0):
                    bounds_ok[pump] = True
                else:
                    bounds_ok[pump] = False
                    logger.info("Bounds on discharge of {} are too close together for optimization. No goals will be added for this pump.".format(pump))
        return bounds_ok

    def path_goals(self):
        # Read the goals from other mixins.
        g = super().path_goals()

        # Read bounds
        bounds = self.bounds()

        # Read parameters
        parameters = self.parameters(0)

        # Get dict to validate pump bounds
        bounds_ok = self.bounds_ok()

        #### Priority 1 ####
        # Wind related goals: keep waterlevel of nodes in the network and inside waterlevel of pumps within range

        # Set a target min and on the pump downstream level to make the pumps
        # reduce / increase pumping when the downstream level drops/rises.
        if self.use_pump_Hdown_range_goal:
            logger.info("(prio 1) Wind related goal: Adding goals for the pump downstream water levels:")
            for pump in self.boezem_pumps:
                try:
                    lower_level_goal = self.get_timeseries(pump + '_Hdown_min').values
                    upper_level_goal = self.get_timeseries(pump + '_Hdown_max').values
                    #TODO: change this to read the actual bounds on the pump Hdown
                    lower_level_goal[lower_level_goal < self._H_state_bounds[0]] = self._H_state_bounds[0]
                    upper_level_goal[upper_level_goal > self._H_state_bounds[1]] = self._H_state_bounds[1]
                    lower_level_goal = Timeseries(self.times(), lower_level_goal)
                    upper_level_goal = Timeseries(self.times(), upper_level_goal)
                except KeyError:
                    # Skip if missing
                    logger.info("          - Did not find a Hdown_min and Hdown_max timeseries for pump {}. Skipping.".format(pump))
                    continue
                else:
                    if bounds_ok[pump]:
                        # used to be with function_range = -5.0, 5.0
                        logger.info("          - Found a Hdown_min and Hdown_max timeseries for pump {}. Adding goal.".format(pump))
                        g.append(WaterLevelRangeGoal(self, pump + '.HQDown.H', self._H_state_bounds, lower_level_goal, upper_level_goal, priority=1, weight=0.5))
                    else:
                        logger.warning("       - Did not add goal for the pump downstream water level for {}, because the bounds on pump discharge are too close together.".format(pump))

        # Set a target min and on several nodes in the network
        if self.use_Hnode_range_goal:
            logger.info("(prio 1) Wind related goal: Adding a goal for the water level range at nodes:")
            for sn in self.storage_nodes:
                try:
                    target_max = parameters[sn + '_H_MAX']
                    target_min = parameters[sn + '_H_MIN']
                except KeyError:
                    # Skip if missing
                    continue
                else:
                    logger.info("          - Adding a goal for the water level range of " + sn)
                    g.append(WaterLevelRangeGoal(self, sn + '.HQ.H', bounds[sn + '.HQ.H'], target_min, target_max, priority=1, sol_path_timeseries_id=sn + '_H_range_violation'))


        #### Priority 2 ####

        # Try to get the water level within the critical range + End of the World H_setpoint goal
        if self.use_H_range_goal:
            logger.info("(prio 2) Adding a goal for the RBP water level range.")
            g.append(WaterLevelRangeGoal(self, 'RBP_H', self._H_state_bounds, self.ondergrens_RBP, self.bovengrens_RBP, priority=2, sol_path_timeseries_id='RBP_H_crit_range_violation'))

        if self.use_H_setpoint_goal:
            RBP_setpoint_goal = self.get_timeseries('RBP_Hstreef')
            logger.info("(prio 2) Adding a goal for water level setpoint (End-of-World).")
            g.append(WaterLevelSetpointGoal(self, 'RBP_H', self._H_state_bounds, RBP_setpoint_goal, self.streefpeil_RBP_band, priority=2))

        #### Priority 20 (chloride related goals) ####

        if self.use_Cl_flushing_goal:
            # Set a minimum request on the corresponding flushing pump
            # TODO: get flushing_pump from polder parameter config?
            #       For now, this is the only flushing pump
            flushing_pump = 'GEM_Gouda'
            state = 'GEM_Gouda_Qoutput'  # represents the sum of the Gouda pumps
            try:
                # TODO: get timeseries name from polder parameter config?
                req_ts_flushing = self.get_timeseries(flushing_pump + '_Qreq_flushing')
            except KeyError:
                logger.warning(
                    'No flushing related discharge request '
                    'timeseries found for pump {}.'.format(flushing_pump))
            else:
                req_len_flushing = np.sum(np.isfinite(req_ts_flushing.values))
                logger.info(
                    '(prio 20) Flushing related discharge request found for pump '
                    '{} at {} timesteps'.format(flushing_pump, req_len_flushing))
                if req_len_flushing > 0:
                    # The timeseries has at least one finite value
                    b1 = bounds[flushing_pump + '_Q']
                    b2 = bounds[flushing_pump + '_inflow_Q']
                    f_range = np.min(b1[0].values) - 1.0, np.max(b2[1].values) + 1.0
                    g.append(
                        DischargeGoal(self,
                            state=state, target_min=req_ts_flushing,
                            order=1.0, function_range=f_range, priority=20))

        # Add goals to get chloride levels below thresholds
        if self.use_Cl_range_goal:
            # Don't use the Boskoop storage node, this chloride max needs some special work.
            # for sn in [sn for sn in self.storage_nodes if sn not in ['ML_Boskoop']]:
            for sn in self.storage_nodes:
                try:
                    target_max = self.get_timeseries(sn + '_Cmax')
                except KeyError:
                    # Skip if missing
                    continue
                else:
                    logger.info('(prio 20) Chloride range goal found for node {}.'.format(sn))
                    relaxation = np.nanmax(target_max.values) * 0.2
                    g.append(ChlorideRangeGoal(self, sn + '.HQ.C', target_max, relaxation=relaxation, priority=20))

        #### Priority 30 ####

        # Add goals to minimize use of emergency pumps
        if self.use_Q_emergency_pump_minimize_goal:
            logger.info("It is possible that emergency pumps will be used. Adding a goal for the minimization of emergency pump use, so that they are only used when it's critical.")
            # TODO: this goal should not apply when the emergency pumps are used to pump water
            # out of the retention area.
            for pump in self.emergency_pumps:
                # Get bounds
                b = bounds[pump + '_Q']

                # This goal is triggered off and on using two inputs:
                # 1) Setting self.use_Q_emergency_pump_minimize_goal to true (via parameters/scenario)
                # 2) Providing timeseries bounds that have more than min_max_bound_buffer space
                #    between the upper and lower bound at at least one timestep in the series
                #    (bounds_ok[pump]=True)

                if bounds_ok[pump]:
                    # Both above conditions hold, so this goal is applied:
                    f_lb = np.amin(b[0].values) - 1.0
                    f_ub = np.amax(b[1].values) + 1.0
                    f_nom = np.max(np.abs([f_ub, f_lb])) / 10
                    logger.info('(prio 30) Emergency pump {} is available.'.format(pump))
                    g.append(MinimizeQpumpGoal(self, pump_var=pump + '_Q', function_range = [f_lb, f_ub], priority=30))
                else:
                    logger.warning("Did not add goal for the minimization of emergency pump {} use, as the bounds are too close together.".format(pump))

        #### Priority 40 / 41 ####

        if self.use_Q_pump_req_goal:
            logger.info("(prio 40) Adding a goal for the general pump discharge requests:")

            for pump in self.pumps:
                # Check if there is a request timeseries
                try:
                    req_ts = self.get_timeseries(pump + '_Qreq')
                except KeyError:
                    logger.info("          - No request timeseries found for pump {}.".format(pump))
                    continue
                else:
                    # A request timeseries was found
                    req_len = len([r for r in req_ts.values if np.isfinite(r)])
                    logger.info("          - Request found for pump {} at {} timesteps.".format(pump, req_len))
                    if req_len:
                        if pump == 'GEM_Gouda':
                            state = 'GEM_Gouda_Qoutput'  # represents the sum of the Gouda pumps
                        else:
                            state = pump + '_Q'
                        # The timeseries has at least one finite value
                        if bounds_ok[pump]:
                            b = bounds[pump + '_Q']
                            f_range = np.nanmin(b[0].values) - 1.0, np.nanmax(b[1].values) + 1.0
                            g.append(
                                DischargeGoal(self,
                                    state=state, target_min=req_ts, target_max=req_ts,
                                    target_relaxation=0.1, function_range=f_range, priority=40))
                        else:
                            logger.warning(
                                '          - Did not add general discharge request for {}, because the '
                                'bounds on pump discharge are too close together.'.format(pump))

        if self.use_katwijk_tide_goals:
            pump = 'GEM_Katwijk'
            # Note: Pump direction is negative, so a max_increase is a minimum for the goal

            q_req_ts = self.get_timeseries('GEM_Katwijk_Qreq_tide')
            #TODO: target max is redundant when this req is 0.0
            if np.any(np.isfinite(q_req_ts.values)) and bounds_ok[pump]:
                b = bounds[pump + '_Q']
                f_range = np.nanmin(b[0].values) - 1.0, np.nanmax(b[1].values) + 1.0
                logger.info("(prio 40) Adding tide-related goals for GEM_Katwijk_Q: Flow Request 0 at rising edge.")
                g.append(
                    DischargeGoal(self,
                        state=pump + '_Q', target_min=q_req_ts, target_max=q_req_ts,
                        target_relaxation=0.1, order=1.0, function_range=f_range, priority=40))
            else:
                logger.warning("Not adding tide-related Request goal for GEM_Katwijk_Q")

            max_increase = self.get_timeseries('GEM_Katwijk_derQmax_tide')
            threshold = parameters.get('GEM_Katwijk_threshold', None)
            if np.any(np.isfinite(max_increase.values)) and bounds_ok[pump]:
                logger.info("(prio 41) Adding tide-related goals for GEM_Katwijk_Q: Rate Of Change Request 0, above threshold of {}m from Hlow of tide.".format(threshold))
                g.append(
                    QP_ROC_Goal(self, state=pump + '_Q', priority=41, max_increase=max_increase))
            else:
                logger.warning("Not adding tide-related ROC goal for GEM_Katwijk_Q")

        return g

    def goals(self):
        # Read the goals from other mixins.
        g = super().goals()

        # Get dict to validate pump bounds
        bounds_ok = self.bounds_ok()

        #### Priorities 50-53 ####

        # Set goals for the minimization of pump costs
        logger.info("Adding a goal for the minimization of pump cost (discharge times unit price):")
        for pump in self.boezem_pumps:
            pump_var = pump + '_Q'
            sign = self.pump_sign[pump]
            price_var = pump + '_UnitPrice_m3s'
            pump_prio = self.pump_priorities[pump]
            relaxation = 0.0

            # Check if there is a price signal
            try:
                price = self.get_timeseries(price_var)
            except KeyError:
                logger.warning("       - (prio {}) No unit price found for {}. Using default value of 1.0".format(pump_prio, pump))
                price = Timeseries(self.times(), np.full_like(self.times(), 1.0))

            if bounds_ok[pump]:
                # TODO: Snap really small prices to zero?
                nonzero_prices = price.values[price.values != 0.0]
                if len(nonzero_prices) > 0:
                    # Calculate a "nominal" function value
                    nominal_price = np.mean(price.values)
                    nominal_flow_rate = 1.0 #m^3/s TODO: find a better way to estimate this?
                    function_nominal = nominal_flow_rate * nominal_price
                else:
                    # Handle case where there are no nonzero prices
                    logger.warning("       - Did not add pump cost minimization goal for {} because all prices were zero.".format(pump))
                    continue
                # Append goal
                logger.info("          - (prio {}) Request found for pump {}.".format(pump_prio, pump))
                goal = MinimizePumpCost(pump_var=pump_var, pump_sign=sign, price_var=price_var, function_nominal=function_nominal,
                                              priority=pump_prio, relaxation=relaxation, sol_path_timeseries_id=pump+'_minimized_cost')
                self.pump_cost_goals.append(goal)
                self.pump_priorities_active.add(pump_prio)
            else:
                logger.warning("       - Did not add pump cost minimization goal for {}, because the bounds on pump discharge are too close together.".format(pump))

        if self.use_cost_pump_minimize_goal_var:
            g.extend(self.pump_cost_goals)

        return g

    def homotopy_options(self):
        options = super().homotopy_options()
        # For now, just hard code this value TODO: make it a parameter without recursion errors
        # options['delta_theta_0'] = 1.0
        options['delta_theta_0'] = 0.5
        options['delta_theta_min'] = 0.05
        return options

    def goal_programming_options(self):
        options = super().goal_programming_options()
        options['constraint_relaxation'] = 1e-5
        # options['fix_minimized_values'] = False
        return options

    def solver_options(self):
        options = super().solver_options()
        solver = options['solver']
        options[solver]['linear_solver'] = 'ma97'
        options[solver]['max_iter'] = 500  # 3000
        # options[solver]['mu_strategy'] = 'adaptive'
        # options[solver]['acceptable_iter'] = 5
        options[solver]['tol'] = 1e-5
        options[solver]['acceptable_tol'] = 1e-4
        options['expand'] = True
        # Options[solver] from Jorn
        options[solver]['bound_relax_factor'] = 1e-9
        options[solver]['bound_push'] = 1e-8
        options[solver]['bound_frac'] = 1e-8
        options[solver]['honor_original_bounds'] = 'no'
        # TODO: this option could result in performance improvement, but
        # has been observed to result in IPOPT failing in some cases.
        # if self._post_first_priority:
        #     options['mu_init'] = 1e-8
        return options

    @property
    def output_variables(self):
        variables = super().output_variables
        additional_output = []
        additional_output.extend([
            MX.sym(variable)
            for variable in self.additional_output])
        variables = list(set().union(variables, additional_output))
        return variables

    def priority_started(self, priority):
        super().priority_started(priority)

        # Return if theta is not 1
        if self.h_th != 1.0:
            return

        # Determine pump analysis strategy
        # Options are 'serial', 'subprocess', 'none'
        # logger.info('pumping analysis - len(self.pump_cost_goals) + len(self.pump_cost_path_goals): {}'.format(str(len(self.pump_cost_goals) + len(self.pump_cost_path_goals))))
        if len(self.pump_cost_goals) + len(self.pump_cost_path_goals) > 0:
            if self.pumping_analysis == 'none':
                logger.warning('The pumping analysis mode is turned off. Impact timeseries will NOT be generated')
            self.pump_analysis_processes = []
        else:
            self.pumping_analysis = 'none'
            logger.warning('The pumping analysis mode is not possible because no sensible pump cost minimization goals can be made. Impact timeseries will NOT be generated')

        if self.pumping_analysis not in ['subprocess', 'serial']:
            return

        if getattr(self, 'previous_priority', None) is None:
            # Initialize pump solution path file
            with open(os.path.join(OUTPUT_FOLDER, 'pump_costs.csv'), 'w') as csv_file:
                writer = csv.writer(csv_file)
                writer.writerow(['priority'] + sorted(self.pumps))
                # we store the baseline costs of 0.0 at a priority number less then the lowest priority
                baseline_cost_prio = priority - 1000
                writer.writerow([baseline_cost_prio] + [0.0] * len(self.pumps))

            # Update previous_priority
            self.previous_priority = priority
            return

        if priority < min(self.pump_priorities_active):
            # Construct pickleable constraints list
            superproblem_constraints = [
                (dill.dumps(constr.function), constr.min, constr.max)
                for key, [constr] in self._GoalProgrammingMixin__subproblem_constraints[0].items()
            ]
            superproblem_path_constraints = [
                (dill.dumps(constr.function), constr.min, constr.max)
                for key, [constr] in self._GoalProgrammingMixin__subproblem_path_constraints[0].items()
            ]

            # Aggregate and deepcopy relevant attributes
            attribs = copy.deepcopy({
                "this_priority": self.previous_priority,
                "previous_prio_results": self.extract_results(),
                "superproblem_constraints": superproblem_constraints,
                "superproblem_path_constraints": superproblem_path_constraints,
                "_full_presim_results": FULL_PRESIM_RESULTS,
                "pump_cost_goals": self.pump_cost_goals,
                "pump_cost_path_goals": self.pump_cost_path_goals
            })

            previous_log_level = logger.getEffectiveLevel()
            logger.info('Running pump analysis')

            if self.pumping_analysis == 'serial':
                execute_pumping_analysis(attribs)
            elif self.pumping_analysis == 'subprocess':
                p = Process(target=execute_pumping_analysis, args=(attribs,))
                p.start()
                self.pump_analysis_processes.append(p)

            # Store prio for use in next run
            self.previous_priority = priority
            logger.setLevel(previous_log_level)

    # We want to print some information about our goal programming problem,
    # like how the results of our highest priority goal (water level) are
    # remembered. The other information we want to see is how our lower
    # priority goals progress.
    # We store the useful numbers temporarily, and print information at the
    # end of our run (see post() method below). This is to avoid our
    # information being lost in the output of IPOPT.
    def priority_completed(self, priority):

        super().priority_completed(priority)

        # Extract results
        results = self.extract_results()
        # Assign GEM_Gouda_inflow_Q results to GEM_Gouda_Q
        results['GEM_Gouda_Q'] += results['GEM_Gouda_inflow_Q']
        results['GEM_Gouda_inflow_Q'][:] = 0.0

        # Append Solution Path Values
        this_theta = self.sol_path_dict.get(self.h_th, {})
        this_theta[priority] = sol_path_entry(self, results=results, solver_stats=self.solver_stats)
        self.sol_path_dict[self.h_th] = this_theta

        if self.output_sol_path_after_each_prio:
            d = self.sol_path_dict
            if len(d) > 0:
                sol_path_df = pd.concat(map(pd.DataFrame, d.values()), keys=d.keys()).stack().unstack(1)
                sol_path_df.index.rename(('theta','priority'), inplace=True)
                sol_path_df.to_csv(os.path.join(OUTPUT_FOLDER, 'sol_path.csv'))

    def post(self):

        # Determine if final run was successful
        successfull_return_strings = {'Solve_Succeeded', 'Solved_To_Acceptable_Level'}
        success = self.solver_stats['return_status'] in successfull_return_strings

        # Extract results
        results = self.extract_results()

        for reach in self.closed_reaches:
            # There are no results for Q[2] at closed reaches, because
            # it is considered a parameter with value 0. Adding a 'fake'
            # series to fill the output timeseries.
            results[reach + '.Q[2]'] = np.full_like(self.times(), 0)

        # Read parameters
        parameters = self.parameters(0)

        ### Output Timeseries providing information on the impact of each pump at each prio ###

        # Mapping of priorities to categories
        # missing wind: its hard to isolate from quantity
        pump_analysis_categories = ['water_quality', 'water_quantity', 'user_constraints', 'energy_cost']
        priority_categories = {
            0:  'water_quality',
            1:  'water_quantity',
            2:  'water_quantity',
            10: 'water_quantity',
            20: 'water_quality',
            30: 'water_quantity',
            40: 'user_constraints',
            41: 'water_quality'}

        # Initialize results dict
        pump_analysis_results = {pump: {k: np.nan for k in pump_analysis_categories} for pump in self.boezem_pumps}

        # Get the net cost of each pump. We simplify by taking the absolute
        # value, but it should be accurate enough for the purpose of this
        # analysis
        pump_net_costs = {}
        total_time = self.times()[-1] - self.times()[0]
        for pump in self.boezem_pumps:
            try:
                pump_net_costs[pump] = self.get_timeseries(pump + '_minimized_cost').values[0] * total_time
            except KeyError:
                pump_net_costs[pump] = 0.0

        # Merge GEM_Gouda_inflow cost and flow results into GEM_Gouda
        pump_net_costs['GEM_Gouda'] += pump_net_costs['GEM_Gouda_inflow']
        pump_net_costs['GEM_Gouda_inflow'] = 0.0
        # A quick check to see if IPOPT was able to make sure both pumps did not run at the same time
        # TODO: change to warning?
        # Possible cause: improper goal configuration that constrains only one of these, or poorly scaled pump minimization
        assertion_tolerance = 0.1
        assert np.all(np.abs(results['GEM_Gouda_Q'] * results['GEM_Gouda_inflow_Q']) < assertion_tolerance)
        results['GEM_Gouda_Q'] += results['GEM_Gouda_inflow_Q']
        results['GEM_Gouda_inflow_Q'][:] = 0.0

        total_cost = sum(pump_net_costs.values())

        # Save pump cost contribution- these are directly proportional to final total cost
        for pump in self.boezem_pumps:
            pump_analysis_results[pump]['energy_cost'] = pump_net_costs[pump]

        # First check if pump analysis was enabled
        if self.pumping_analysis in ['subprocess', 'serial']:

            # Wait until all subprocesses have stopped
            for p in self.pump_analysis_processes:
                p.join()

            # Determine proportional contribution of each pump
            proportion_contribution = {}
            for pump in self.boezem_pumps:
                try:
                    proportion_contribution.update({pump: pump_net_costs[pump] / total_cost})
                except ZeroDivisionError:
                    proportion_contribution.update({pump: 0.0})

            # Dataframe contains the minimized pump cost at non-pump priorities, with each pump separately listed
            df = pd.read_csv(os.path.join(OUTPUT_FOLDER, 'pump_costs.csv'))

            # Make list of priorities. Pump priorities are excluded, and final
            # results of main optimization correspond to the last priority before
            # pump priorities. This is because pump priorities are minimization
            # goals applied to the decision variables, so they can't constrain
            # the objective but rather determine its final solution
            priorities = sorted([int(p) for p in df['priority']] + [self.previous_priority])

            # Configure dataframe
            df = df.set_index('priority')
            df = df[self.boezem_pumps]

            # Set results of main optimization as the value at previous priority
            df.loc[self.previous_priority] = pump_net_costs

            # Loop over priorities and set influence contributions. The final
            # contribution of a pump to a given analysis category is proportional
            # to the final results. However, we use the intermediate results to
            # determine the proportion of the final results that can be attributed
            # to each category. We we so this by adding up the cost that each goal
            # caused on each pump, but scaling by that pump's proportional
            # contribution to the final results. This is not perfect, but given
            # the limited data and runtime constraints it can give a useful
            # measure of the influence of each analysis category on each pump and
            # their relative impact on the final results
            for prev_prio, prio in zip(priorities[:-1], priorities[1:]):
                category = priority_categories[prio]
                delta_cost = np.sum(df.loc[prio]) - np.sum(df.loc[prev_prio])
                for pump in self.boezem_pumps:
                    if np.isnan(pump_analysis_results[pump][category]):
                        pump_analysis_results[pump][category] = delta_cost * proportion_contribution[pump]
                    else:
                        pump_analysis_results[pump][category] += delta_cost * proportion_contribution[pump]

        # Set analysis results.
        empty_array = np.full_like(self.times(),np.nan)
        for pump, analysis_category in itertools.product(self.boezem_pumps, pump_analysis_categories):
            result = pump_analysis_results[pump][analysis_category]
            array = empty_array.copy()
            logger.info('pump {}, analysis_category: {} - total_cost = {}, result = {}'.format(pump,analysis_category,str(total_cost),str(result)))
            if (not np.isnan(result)) and (total_cost != 0.0) and (self.pumping_analysis in ['subprocess', 'serial']):
                # If result is real, set the first element to the scaled result
                # We scale the impact number to be an integer from 1 to 3, rounding up for (0.0, 1.0]
                # 0: no information/no active goals/no cost
                # 1: low impact
                # 2: medium impact
                # 3: high impact
                scaled_result = result * 3.0 / total_cost
                if 0.0 < scaled_result <= 1.0:
                    scaled_result = 1.0
                elif 1.0 < scaled_result <= 2.0:
                    scaled_result = 2.0
                else:
                    scaled_result = 3.0
                array[0] = scaled_result
            else:
                # Always return a value for the impact timeseries - if there is no information/no active goals (NaN), return a 0.0
                array[0] = 0.0
            self.set_timeseries('_'.join((pump, analysis_category, 'impact')), Timeseries(self.times(), array))

        # only run sim if optimization was successful
        if success:
            # Gain of i/d relates to time to correct integral error (Ti) and
            # time to correct derivative error (Td). FEWS supplies these in h.
            #   I_gain should then be set to 1/(Ti)
            #   D_gain should then be set to (Td)
            # Either way, delta t is taken into account in the PID-controller,
            # so the gain parameters of I and D should take that into account.
            # With I_gain = 1/dt, D_gain = dt they both 'weigh' the same.
            p_i_d_gain = (parameters['GAINp'], 1.0/(3600.0*parameters['Ti']), 3600*parameters['Td'])

            # Calculate best discrete setpoints
            discrete_pump_setpoints = {}
            extra_outputs = {}

            # Add pump analysis ts to extra_outputs
            for pump, analysis_category in itertools.product(self.boezem_pumps, pump_analysis_categories):
                ts_name = '_'.join((pump, analysis_category, 'impact'))
                extra_outputs[ts_name] = self.get_timeseries(ts_name).values

            for pump in self.pumps:
                # Get min and max pump TS (None defaults to min and max pump stage)
                if self.adjust_timesteps:
                    try:
                        m = self.equidistant_timeseries_import.get(pump + '_Qmax')
                    except KeyError:
                        m = None
                    try:
                        M = self.equidistant_timeseries_import.get(pump + '_Qmin')
                    except KeyError:
                        M = None
                    continous_flow = np.interp(self.equidistant_times, self.timeseries_import_times, results[pump + '_Q'])
                    dt = self.equidistant_deltatime.total_seconds()
                else:
                    try:
                        m = self.get_timeseries(pump + '_Qmax').values
                    except KeyError:
                        m = None
                    try:
                        M = self.get_timeseries(pump + '_Qmin').values
                    except KeyError:
                        M = None
                    continous_flow = results[pump + '_Q']
                    dt = self.timeseries_import.dt.total_seconds()
                if pump == 'GEM_Gouda':
                    # Simulation mode treats the GEM_Gouda pumps as one pump
                    stages = np.vstack([self.pump_stages['GEM_Gouda'], self.pump_stages['GEM_Gouda_inflow'][1:]])
                    M = self.get_timeseries('GEM_Gouda_inflow_Qmin').values
                else:
                    stages = self.pump_stages.get(pump, None)
                if stages is not None:
                    modes, flow_rates = discrete_setpoint_controller(continous_flow, stages,
                                                                     *p_i_d_gain, dt=dt,
                                                                     min_flow=m, max_flow=M)
                else:
                    modes, flow_rates = np.zeros_like(continous_flow), continous_flow
                discrete_pump_setpoints[pump + '_Q'] = flow_rates
                extra_outputs[pump + '_Qoutput_mode'] = modes

            suffix = '_postprocessed'

            attribs = dict(
                export_suffix = suffix,
                discrete_pump_setpoints = discrete_pump_setpoints,
                extra_outputs = extra_outputs,
                super_problem = self
                )

            # Make a special post_simulation class with appropriate attributes
            post_simulation = type('Rijnland_PostSimulation' + suffix, (Rijnland_PostSimulation,), attribs)

            # Run Subproblem
            logger.info('Simulating Model With Pump Discretization Parameters P={}, I={}, D={}.'.format(*p_i_d_gain))
            run_simulation_problem(post_simulation, log_level=logging.INFO)

        # Call super to write output to XML
        super().post()


        # Postprocess results into equidistant steps
        if self.equidistant_export_times and self.adjust_timesteps:
            logger.info('TimestepRescalingMixin: Overwriting output timeseries with interpolated equidistant timesteps.')
            self.timeseries_export._times = self.equidistant_datetimes
            self.timeseries_export.dt = self.equidistant_deltatime

            for var_name, values in self.timeseries_export.items():
                values = np.interp(self.equidistant_times, self.timeseries_import_times, values)
                self.timeseries_export.set(var_name, values)
            for var, vals in results.items():
                results[var] = np.interp(self.equidistant_times, self.timeseries_import_times, vals)

        # Output Solution Path Dataframe as CSV
        d = self.sol_path_dict
        if len(d) > 0:
            sol_path_df = pd.concat(map(pd.DataFrame, d.values()), keys=d.keys()).stack().unstack(1)
            sol_path_df.index.rename(('theta','priority'), inplace=True)
            sol_path_df.to_csv(os.path.join(OUTPUT_FOLDER, 'sol_path.csv'))

        # Get path goals as list
        goal_list = self.path_goals()

        # Generate csv
        process_sol_path_output(goal_list, self.sol_path_dict, OUTPUT_FOLDER)


class Rijnland_Post_Process(Rijnland_base):
    # TODO: remove this class, or make it more robust with regard to files I/O.
    # Now the class will overwrite the original timeseries_export_postprocessed
    # in the output folder.

    initial_node_levels = {}
    initial_node_concentrations = {}

    # Prepare dicts for best discrete setpoints
    discrete_pump_setpoints = {}
    extra_outputs = {}

    suffix = '_postprocessed'

    pi_parameter_config_pid_basename = 'rtcParameterConfig_Boezem-PID'

    basename_import = 'timeseries_import' + '-rerun'
    basename_export = 'timeseries_export'

    def pre(self):
        logger.info("RijnlandPostProcess: Reading input.")

        self._data_config = rtc.DataConfig(INPUT_FOLDER)

        # In the current set-up we assume that the files of a previous run are
        # still present. This because we would otherwise need a lot of changes
        # in FEWS and hopefully this situation is temporary.

        # Copy the file we need to the input folder
        shutil.copy(os.path.join(OUTPUT_FOLDER, self.basename_export + '.xml'), os.path.join(INPUT_FOLDER, self.basename_import + '.xml'))
        try:
            self.timeseries_import = pi.Timeseries(
                self._data_config, INPUT_FOLDER, self.basename_import, binary=False, pi_validate_times=False)
        except IOError:
            raise Exception("PI: {}.xml not found in {}.".format(
                basename_import, OUTPUT_FOLDER))

        for node in self.storage_nodes:
            self.initial_node_levels[node] = self.timeseries_import.get(node + '.HQ.H')[0]
            self.initial_node_concentrations[node] = self.timeseries_import.get(node + '.HQ.C')[0]

        # TODO: relate gain to time to correct integral error (Ti) and time
        # to correct derivative error (Td)?
        #   I_gain should then be set to 1/(Ti)
        #   D_gain should then be set to (Td)
        # Either way, delta t is taken into account in the PID-controller,
        # so the gain parameters of I and D should take that into account.
        # With I_gain = 1/dt, D_gain = dt they both 'weigh' the same.
        p = pi.ParameterConfig(INPUT_FOLDER, self.pi_parameter_config_pid_basename)
        self.p_i_d_gain = (p.get(group_id='Boezem-PID', parameter_id='GAINp'),
                           1.0/(3600.0 * p.get(group_id='Boezem-PID', parameter_id='Ti')),
                           3600.0 * p.get(group_id='Boezem-PID', parameter_id='Td'))

    def apply_post_processing(self):
        logger.info("RijnlandPostProcess: Applying post-processing.")
        for pump in self.pumps:
            continous_flow = self.timeseries_import.get(pump + '_Q')
            dt = self.timeseries_import.dt.total_seconds()
            stages = self.pump_stages.get(pump, None)

            if stages is not None:
                modes, flow_rates = discrete_setpoint_controller(continous_flow, stages, *self.p_i_d_gain, dt=dt)
            else:
                modes, flow_rates = np.zeros_like(continous_flow), continous_flow
            logger.info("pump_Q output is set for pump {}".format(pump))
            self.discrete_pump_setpoints[pump + '_Q'] = flow_rates
            self.extra_outputs[pump + '_Qoutput_mode'] = modes

    def simulate_results(self):
        logger.info("RijnlandPostProcess: Simulating with post-processed results.")
        attribs = dict(
            export_suffix = self.suffix,
            discrete_pump_setpoints = self.discrete_pump_setpoints,
            extra_outputs = self.extra_outputs,
            initial_node_levels = self.initial_node_levels,
            initial_node_concentrations = self.initial_node_concentrations
            )

        # Make a special post_simulation class with appropriate attributes
        post_simulation = type('Rijnland_PostSimulation' + self.suffix, (Rijnland_PostSimulation,), attribs)

        # Run Subproblem
        logger.info('Simulating Model With Pump Discretization Parameters P={}, I={}, D={}.'.format(*self.p_i_d_gain))
        run_simulation_problem(post_simulation, log_level=logging.INFO)

        # Copy the results to a new file
        shutil.copy(os.path.join(OUTPUT_FOLDER, self.basename_export + self.suffix + '.xml'), os.path.join(OUTPUT_FOLDER, self.basename_export + self.suffix + '-rerun.xml'))

        # Copy back the original input file
        shutil.copy(os.path.join(INPUT_FOLDER, self.basename_import + '.xml'), os.path.join(OUTPUT_FOLDER, self.basename_export + '.xml'))

    def run(self):
        self.pre()
        if self.timeseries_import.dt != None:
            self.apply_post_processing()
            self.simulate_results()
        else:
            logger.error("RijnlandPostProcess: input file has non-equidistant timeseries. These are not supported.")
            raise Exception


class Rijnland_Full_Presim(Rijnland_base, PIMixin_sim, SimulationProblem):
    model_name = 'PI_Pump_Full_Sim'
    timeseries_import_basename = 'timeseries_import-initialized'

    def pre(self):
        super().pre()
        # We make a timeseries of nans b/c we set the calculated value of Q_actual in the update() method
        nans = np.full_like(self.times(), np.nan)
        for pump in self.pumps:
            try:
                q0 = self.get_timeseries(pump + '_Q0_guess')[0]
                # If q0 is missing, we constrain to a safe value of 0.0
                if np.isnan(q0):
                    q0 = 0.0
                    logger.warning('Pump {} has no value for the first timestep, Q0 will be set to 0.0'.format(pump))
            except KeyError:
                # If there are any KeyErrors, we constrain to a safe value of 0.0
                q0 = 0.0
            q0_series = np.copy(nans)
            q0_series[0] = q0
            self.set_timeseries(pump + '_Q_actual', q0_series)

        self.use_Cl_flushing_goal = self.parameters()['use_Cl_flushing_goal']

    @cached
    def get_output_variables(self):
        output_vars = []
        for reach in self.active_reaches:
            output_vars.append(reach + '.Q[2]')
            output_vars.append(reach + '.HQUp.C')
            output_vars.append(reach + '.HQDown.C')
            output_vars.append(reach + '.HQUp.H')
            output_vars.append(reach + '.HQDown.H')

        for pump in self.pumps:
            output_vars.append(pump + '.HQDown.C')
            output_vars.append(pump + '.HQUp.C')

        for sn in self.storage_nodes:
            output_vars.append(sn + '.V')
            output_vars.append(sn + '.HQ.C')

        return output_vars

    def update(self, dt):
        # Get the time
        time = self.get_current_time()

        for pump in self.pumps:
            q_req = np.nan

            # If there is a flushing request, use it
            if self.use_Cl_flushing_goal:
                try:
                    q_req = self.timeseries_at(pump + '_Qreq_flushing', time)
                except KeyError:
                    q_req = np.nan

            # if no flushing request, look for normal requests
            if np.isnan(q_req):
                try:
                    q_req = self.timeseries_at(pump + '_Qreq', time)
                except KeyError:
                    q_req = np.nan

            # There were no requests, so use the expected flow rate as decided by the PI controllers
            if np.isnan(q_req):
                try:
                    q_req = self.get_var(pump + '_Q_expected').item()
                except KeyError:
                    q_req = 0.0

            logger.debug('Presim: at t={}, {}.Q={}'.format(time, pump, q_req))

            self.set_var(pump + '_Q_actual', q_req)

        # Call the super class so that everything else continues as normal
        super().update(dt)

    def post(self):
        # Make a global variable to store presim results to memory- this saves a reading/writing to disk
        global FULL_PRESIM_RESULTS
        FULL_PRESIM_RESULTS = self.extract_results()


class Rijnland_PostSimulation(Rijnland_base, PIMixin_sim, SimulationProblem):
    '''Post Simulation to Validate Pump Flows'''
    model_name = 'Discrete_Pump_Post_Sim'
    timeseries_import_basename = 'timeseries_import-initialized'

    # Some empty class variables that will be set in subclasses
    discrete_pump_setpoints = {}
    extra_outputs = {}
    export_suffix = None

    def get_output_variables(self):
        output_variables = set(super().get_output_variables())

        # Output the values of all states
        output_variables.update(set(self.get_variables().keys()))

        # To make sure the timeseries export doesn't miss any states because
        # we are looking for the wrong alias name, we extend the output with
        # the additional output variables used by the super_problem
        output_variables.update(set(self.super_problem.additional_output))

        return list(output_variables)

    def pre(self):
        # First specify the name of the output file
        self.timeseries_export_basename = 'timeseries_export' + self.export_suffix

        # Do the usual preprocessing
        super().pre()

        # Set the pump control timeseries
        for input_var, values in self.discrete_pump_setpoints.items():
            self.set_timeseries(input_var, values, unit='m3/s')

        # Set the extra outputs (set_timeseries() adds timeseries to output by default)
        for output_var, values in self.extra_outputs.items():
            self.set_timeseries(output_var, values)

    def post(self):
        super().post()

        # Simulation does not have an extract_results() method
        output_dict = self._PIMixin__output

        # Process results using sol-path criteria
        final_entry = sol_path_entry(self.super_problem, results=output_dict, custom_times=self.timeseries_import_times)

        # Append post-processing results to super_problem sol_path_dict
        self.super_problem.sol_path_dict['post-processing'] = {'post-processing': final_entry}


# Run
if __name__ == '__main__':
    init_start_time = timeit.default_timer()
    start_time = init_start_time

    # Before running any models, calculate polder flow timeseries
    schedule_polder_pumps(INPUT_FOLDER, Rijnland_base.polders)

    # Before running any models, detect any emergency closures
    emergency_closures = EmergencyClosures(model_folder=MODEL_FOLDER,
                                           input_folder=INPUT_FOLDER,
                                           reaches=Rijnland_base.reaches)
    closed_reaches = emergency_closures.get_closed_reaches()

    # To avoid excessive model recompilation, we only recompile model if there
    # was a change in the emergency closures. We store the previous closues in
    # closed_reaches.csv
    try:
        with open(os.path.join(INPUT_FOLDER, 'closed_reaches.csv'), 'r') as csv_file:
            reader = csv.reader(csv_file)
            previous_run_closed_reaches = list(reader)[0]
    except IOError:
        previous_run_closed_reaches = []

    if set(previous_run_closed_reaches) != set(closed_reaches):
        logger.warning('The set of closed reaches have changed. Model will be updated and recompiled.')
        emergency_closures.update_model()
        emergency_closures.update_csv()

    try:
        general = pi.ParameterConfig(INPUT_FOLDER, 'rtcParameterConfig')
        run_type = general.get('General', 'run_type').lower()
    except IOError:
        run_type = 'optimization'

    if run_type == 'optimization':

        # Initialization module to determine consistant initial conditions
        import initialize_states
        # Generate timeseries_import-preinit.xml
        initialize_states.gen_preinit_timeseries()
        # Generate timeseries_import-initialized.xml
        run_optimization_problem(initialize_states.RijnlandInitializeStates)

        # Use Simulation to seed starting values and store in timeseries_import-full_initialization.xml
        run_simulation_problem(Rijnland_Full_Presim, log_level=logging.INFO)

        init_run_time = timeit.default_timer() - start_time
        main_opt_start_time = timeit.default_timer()

        # Run the main problem using the initialized timeseries and presim output
        run_optimization_problem(Rijnland_Optimization, log_level=logging.INFO)

    elif run_type == 'post-processing':
        logger.info("Only running post-processing and simulation.")

        init_run_time = timeit.default_timer() - start_time
        main_opt_start_time = timeit.default_timer()

        post = Rijnland_Post_Process()
        post.run()

    # Report Execution Times
    print('\nExecution time: {:.0f}:{:2.0f}'.format(*divmod(timeit.default_timer() - start_time, 60)))
    print('  Init: {:.0f}:{:2.0f}'.format(*divmod(init_run_time, 60)))
    print('  Main: {:.0f}:{:2.0f}'.format(*divmod(timeit.default_timer() - main_opt_start_time, 60)))
