import xml.etree.cElementTree as ET
import rtctools.data.pi as pi
import os, sys
import csv

import logging
logger = logging.getLogger("rtctools")


class EmergencyClosures:
    sub_sub_model_src = 'Network_orig'
    sub_sub_model_dst = 'Network'

    branch_def_orig = 'LocalHomotopicLinear'
    branch_def_new = 'LocalHomotopicLinearClosedBranch'

    pi_emergency_closure_config_basename = 'rtcParameterConfig_Noodwaterkeringen'

    # Prepare dicts for additional settings, will be filled later
    emergency_closure_reaches = []

    def __init__(self, model_folder, input_folder, reaches):

        # Save arguments
        self.input_folder = input_folder
        self.model_folder = model_folder
        self.reaches = reaches

        ## Get use of emergency closure structures (noodwaterkeringen)
        # Replicating part of pi.pyx, to avoid abusing the API and to
        # be able to loop through the file more efficiently.
        file_path = os.path.join(self.input_folder, self.pi_emergency_closure_config_basename + '.xml')
        # Read file
        try:
            tree = ET.parse(file_path)
        except FileNotFoundError:
            logger.info("EmergencyClosures: {} not found. Assuming no Emergency Closures.".format(file_path))
            self.emergency_closure_reaches = []
        else:
            logger.info("EmergencyClosures: Reading {} to determine which reaches have an emergency closure.".format(file_path))

            xml_root = tree.getroot()

            ns = {'fews': 'http://www.wldelft.nl/fews',
                  'pi': 'http://www.wldelft.nl/fews/PI'}
            # Get and loop over the groups (pairs of branch/emergency closure boolean)
            groups = xml_root.findall("pi:group", ns)
            for group in groups:
                # Read reach id
                el = group.find("pi:parameter[@id='{}']".format('RTC_BRANCH_ID'), ns)
                if el is None:
                    logger.warning("EmergencyClosures: Missing 'RTC_BRANCH_ID' for a group in {}. Ignoring group.".format(file_path))
                else:
                    for child in el: # according to the xsd schema, there should be only one child
                        reach = child.text

                    if reach not in self.reaches:
                        logger.warning("EmergencyClosures: Reach {} referenced in 'RTC_BRANCH_ID' for a group in {} not present in this class. Ignoring group.".format(reach, file_path))
                        continue # go to next group

                    # If reach id is found, read use of emergency closure boolean
                    el = group.find("pi:parameter[@id='{}']".format('INZET_NWK'), ns)
                    if el is not None:
                        for child in el: # according to the xsd schema, there should be only one child
                            emergency_closure = child.text

                        if emergency_closure.lower() == 'true':
                            # Branch id is present in this part of the if-tree, so
                            # add setting to ordered dict.
                            self.emergency_closure_reaches.append(reach)
                        elif emergency_closure.lower() == 'false':
                            pass
                        else:
                            logger.warning("EmergencyClosures: Setting 'INZET_NWK' in {} for reach {} contains non-boolean value. Ignoring it.".format(file_path, reach))

    def update_model(self):
        ## Switch model elements of reaches with emergency closure
        src = open(os.path.join(self.model_folder, self.sub_sub_model_src + '.mo'), 'r')
        dst = open(os.path.join(self.model_folder, self.sub_sub_model_dst + '.mo'), 'w')

        closed_reaches = self.emergency_closure_reaches

        logger.info("EmergencyClosures: Switching reaches with emergency closure to closed branch model elements.")
        for line in src:
            if line.lstrip().startswith("model {}".format(self.sub_sub_model_src)) or \
                    line.lstrip().startswith("end {}".format(self.sub_sub_model_src)):
                line = line.replace(self.sub_sub_model_src, self.sub_sub_model_dst)
            elif line.lstrip().startswith('Deltares.ChannelFlow.Hydraulic.Branches.HomotopicLinear ') and any((r + '(' in line for r in closed_reaches)):
                match_str = next((r for r in closed_reaches if r in line))
                logger.info("EmergencyClosures: Removing channel element {}.".format(match_str))
                # Make the line a comment
                line = line.replace('Deltares.ChannelFlow.Hydraulic.Branches.HomotopicLinear', '//Deltares.ChannelFlow.Hydraulic.Branches.HomotopicLinear')
            elif line.lstrip().startswith('connect(') and any((r + '.HQ' in line for r in closed_reaches)):
                # Note: this matches HQUp and HQDown
                logger.info("EmergencyClosures: Removing {}).".format(line.lstrip().split(')')[0]))
                # Make the line a comment
                line = line.replace('connect(', '//connect(')
            dst.write(line)
        src.close()
        dst.close()

    def get_closed_reaches(self):
        return self.emergency_closure_reaches

    def update_csv(self, fname='closed_reaches'):
        with open(os.path.join(self.input_folder, fname + '.csv'), 'w') as csv_file:
            writer = csv.writer(csv_file, quoting=csv.QUOTE_ALL)
            writer.writerow(self.emergency_closure_reaches)
