from rtctools.optimization.collocated_integrated_optimization_problem \
    import CollocatedIntegratedOptimizationProblem
from rtctools.optimization.goal_programming_mixin import GoalProgrammingMixin, Goal
from rtctools.optimization.initial_state_estimation_mixin import InitialStateEstimationMixin
from rtctools.optimization.homotopy_mixin import HomotopyMixin
from rtctools.optimization.modelica_mixin import ModelicaMixin
from rtctools.optimization.pi_mixin import PIMixin
from rtctools.util import run_optimization_problem
import rtctools.data.pi as pi

# Import a useful data class from the main optimization script
from Rijnland_Boezem_Optimization import Rijnland_base
from utilities import snap_to_discrete

import logging
logger = logging.getLogger("rtctools")

import numpy as np
import xml.etree.cElementTree as ET
import os, sys
from shutil import copyfile

# First make a timeseries_import-preinit.xml with only two collocation points
def gen_preinit_timeseries():
    path_xml = os.path.join(sys.path[0], '..', 'input', 'timeseries_import-polder_scheduling.xml')
    tree = ET.parse(path_xml)
    xml_root = tree.getroot()
    for series in xml_root.iterfind('.//pi:series', pi.ns):
        events = series.findall('pi:event', pi.ns)
        for index, event in enumerate(events):
            if index == 1:
                endDate = series.find('pi:header/pi:endDate', pi.ns)
                endDate.set('date', event.get('date'))
                endDate.set('time', event.get('time'))
            elif index > 1:
                series.remove(event)
    tree.write(path_xml.replace('polder_scheduling.xml', 'preinit.xml'))


class RijnlandInitializeStates(Rijnland_base, InitialStateEstimationMixin, HomotopyMixin, GoalProgrammingMixin,
                               PIMixin, ModelicaMixin, CollocatedIntegratedOptimizationProblem):
    model_name = 'initial_states'
    timeseries_import_basename = 'timeseries_import-preinit'

    def initial_state_measurements(self):
        pairs = super().initial_state_measurements()
        for n in self.storage_nodes:
            pairs.append([n + '.HQ.C', n + '_C0_guess'])
        for n in self.measured_nodes:
            pairs.append([n + '.HQ.H', n + '_H0_guess'])
        return pairs

    def initial_state_smoothing_pairs(self):
        pairs = super(RijnlandInitializeStates, self).initial_state_smoothing_pairs()
        for reach in self.active_reaches:
            pairs.append([reach + '.HQUp.H', reach + '.HQDown.H'])
        return pairs

    def homotopy_options(self):
        options = super().homotopy_options()
        options['delta_theta_0'] = 0.5
        return options

    def solver_options(self):
        options = super().solver_options()
        solver = options['solver']
        options['expand'] = True
        options[solver]['linear_solver'] = 'ma97'
        #options['tol'] = 1e-5
        #options['acceptable_tol'] = 1e-4
        return options

    def path_constraints(self, ensemble_member):
        c = super().path_constraints(ensemble_member)
        # Constrain the initial pump flow rate. If not provided, then zero
        # Note: shouldn't influence the model when the discretization is backwards Euler
        for pump in self.pumps:
            try:
                q0 = self.get_timeseries(pump + '_Q0_guess').values[0]
                # Ensure initial condition is within bounds
                m = self.get_timeseries(pump + '_Qmax').values[0]
                M = self.get_timeseries(pump + '_Qmin').values[0]
                q0 = float(q0 if q0 < M else M)
                q0 = float(q0 if q0 > m else m)
            except KeyError:
                # If there are any KeyErrors, we constrain to a safe value of 0.0
                q0 = 0.0
            c.append((self.state(pump + '_Q'), q0, q0))
        return c

    def constraints(self, ensemble_member):
        c = super().constraints(ensemble_member)
        for reach in self.active_reaches:
            c.append((self.state_at(reach + '.Q[1]', 0.0) - self.state_at(reach + '.Q[2]', 0.0), 0.0, 0.0))
            c.append((self.state_at(reach + '.Q[2]', 0.0) - self.state_at(reach + '.Q[3]', 0.0), 0.0, 0.0))
            c.append((self.der_at(reach + '.Q[2]', 0.0), 0.0, 0.0))
        return c

    def pre(self):
        logger.info('RijnlandInitializeStates: Using InitialStateEstimationMixin to find initial conditions')
        super().pre()

        # Set up inputs for GEM_Gouda_inflow (this is because FEWS only knows about one pump at Gouda)
        pump, pump_inflow = 'GEM_Gouda', 'GEM_Gouda_inflow'
        copied_timeseries = ['_Qmax', '_Qmin', '_Q0_guess']
        for ts in copied_timeseries:
            try:
                self.set_timeseries(pump_inflow + ts, self.get_timeseries(pump + ts))
            except KeyError:
                continue

        self._HomotopyMixin__theta = 0.0 # a quick hack to avoid attribute error when calling parameters()
        parameters = self.parameters(0)

        # Adjust pump min and max timeseries, as they must respect bounds from timeseries,
        # parameters, and discrete settings. we also call this for loop in the main opt problem
        for pump in self.pumps:
            # The highest possible pump flow is stored as a negative number in
            # FEWS, but called 'max'. Here we have to use that as the lower bound.
            m = self.get_timeseries(pump + '_Qmax')

            # The lowest possible pump flow is stored as a positive number in
            # FEWS, but called 'min'. Here we have to use that as the upper bound.
            M = self.get_timeseries(pump + '_Qmin')

            try:
                # First check if param exists
                param_min = parameters[pump + '_Q_MAX'] # Q_MAX is actually a min
            except KeyError:
                pass
            else:
                # Adjust bounds
                m.values[m.values < param_min] = param_min
                self.set_timeseries(pump + '_Qmax', m)

            stages = self.pump_stages.get(pump, None)
            if stages is not None:
                stages = stages[:, 1]
                m = snap_to_discrete(m, stages, strategy='up')
                self.set_timeseries(pump + '_Qmax', m)
                M = snap_to_discrete(M, stages, strategy='down')
                self.set_timeseries(pump + '_Qmin', M)

        # Set Polder Inflows If Missing (TODO: Remove)
        for polder in self.polders:
            try:
                self.get_timeseries('PG_{}_Q'.format(polder))
            except KeyError:
                self.set_timeseries('PG_{}_Q'.format(polder), np.full_like(self.times(), 0.0))
            try:
                self.get_timeseries('BUF_{}_C'.format(polder))
            except KeyError:
                self.set_timeseries('BUF_{}_C'.format(polder), np.full_like(self.times(), 0.0))

        # load regime
        import regime_runner
        self.scenario = self.get_timeseries('RBP_optimalisatie_scenario').values[0]
        regime_runner.loadscenario(self.scenario)

    def priority_completed(self, priority):
        if self.parameters(0)['theta'] == 1.0 and priority == -1:
            logger.info('RijnlandInitializeStates: Done Finding Initial States')

    def post(self):

        # Determine if final run was successful
        successfull_return_strings = {'Solve_Succeeded', 'Solved_To_Acceptable_Level'}
        success = self.solver_stats['return_status'] in successfull_return_strings

        if not success:

            # Check and warn if the issue is related to missing guesses (most probable cause)
            for ts_name, values in self._PIMixin__timeseries_import.items():
                if ts_name.endswith(('_H0_guess', '_C0_guess')) and np.isnan(values[0]):
                    logger.error('RijnlandInitializeStates: Timeseries {} contains missing values'. format(ts_name))

            # Log and quit
            logger.error('RTC-Tools failed to find feasible initial conditions. Possible Cause: missing/bad input data')
            sys.exit(0)

        basename_import = 'timeseries_import'

        self.original_timeseries_import = pi.Timeseries(
            self._PIMixin__data_config, self._PIMixin__input_folder, basename_import + '-polder_scheduling',
            binary=self.pi_binary_timeseries, pi_validate_times=self.pi_validate_timeseries)

        self.modified_timeseries_import = pi.Timeseries(
            self._PIMixin__data_config, self._PIMixin__input_folder, basename_import + '-initialized',
            binary=self.pi_binary_timeseries, pi_validate_times=self.pi_validate_timeseries, make_new_file=True)

        results = self.extract_results()

        self.modified_timeseries_import.times = self.original_timeseries_import.times[self.original_timeseries_import.forecast_index:]

        # Write other time settings
        self.modified_timeseries_import.forecast_datetime  = self.original_timeseries_import.forecast_datetime
        self.modified_timeseries_import.dt = self.original_timeseries_import.dt
        self.modified_timeseries_import.timezone = self.original_timeseries_import.timezone

        # Write the ensemble properties for the export file.
        self.modified_timeseries_import.ensemble_size = self.ensemble_size
        self.modified_timeseries_import.contains_ensemble = self.original_timeseries_import.contains_ensemble

        # Iterate over existing timeseries, modifying the guesses to have actual variable names
        for ts_name, values in self.original_timeseries_import.items():
            if ts_name.endswith(('_H0_guess', '_C0_guess')):
                # Handle some states that were not initialized:
                if ts_name in ['Eindk_Gouda_C0_guess', 'Eindk_Halfweg_C0_guess',
                               'Eindk_Katwijk_C0_guess', 'Eindk_Spaarndam_C0_guess']:
                    continue
                var_name = ts_name.replace('_H0_guess', '.HQ.H').replace('_C0_guess', '.HQ.C')
                try:
                    initial_value = np.array([results[var_name][0]])
                except KeyError:
                    logger.warning('RijnlandInitializeStates: ' +
                                   '{} is not in the model. Timeseries {} was ignored.'.format(var_name, ts_name))
                    continue
                logger.debug('Adjusted initial condition for {}: {} -> {}'.format(var_name, values[0], initial_value[0]))
                self.modified_timeseries_import.set(var_name, initial_value, unit=self.original_timeseries_import.get_unit(ts_name))
                logger.debug('RijnlandInitializeStates.post(): RTC-Tools has replaced' +
                             '{} with {} in {}.xml.'.format(ts_name, var_name, basename_import + '-initialized'))
            else:
                try:
                    self._PIMixin__data_config.pi_variable_ids(ts_name)
                except KeyError:
                    logger.error('RijnlandInitializeStates.post(): variable {} has no mapping defined in \
                        rtcDataConfig and cannot be added to {}.xml.'.format(ts_name, basename_import + '-initialized'))
                    continue
                else:
                    self.modified_timeseries_import.set(ts_name, values, unit=self.original_timeseries_import.get_unit(ts_name))

        # Add new timeseries to modified_timeseries_import to save initial conditions
        for reach in self.active_reaches:
            var_name = reach + '.Q[2]'
            # For consistancy with RTC 2.0 implementation, we set initial Q[2] with a value of zero
            initial_value = np.array([0.0])
            # TODO: see if using the calculated initial q will improve performance, e.g:
            # initial_value = np.array([results[var_name][0]])
            self.modified_timeseries_import.set(var_name, initial_value, unit='CMS')
            logger.debug('Adjusted initial condition for {}: {}'.format(var_name, initial_value.item()))

        self.modified_timeseries_import.write()

        # For testing, it may be helpful to run super post(), but in operation it is not used
        # super().post()

if __name__ == '__main__':
    gen_preinit_timeseries()
    run_optimization_problem(RijnlandInitializeStates)#, log_level=logging.DEBUG)
