"""
This script runs right away in the Rijnland workflow. It creates polder
pumping timeseries and GEM_Gouda flushing request timeseries.

input file: ../input/timeseries_import.xml
output_file: ../input/timeseries_import-polder_scheduling.xml
"""

import logging
import numpy as np
import scipy.signal
import rtctools.data.pi as pi
import rtctools.data.rtc as rtc
from rtctools.optimization.timeseries import Timeseries

logger = logging.getLogger("rtctools")


def schedule_polder_pumps(
    input_folder,
    polders,
    basename_import='timeseries_import',
    tide_timeseries='Eindk_Gouda_H',
    tide_peak_timeseries='Eindk_Gouda_Hpeak',
    tide_low_timeseries='Eindk_Gouda_Hlow',
    flushing_pump_timseries_req='GEM_Gouda_Qreq_flushing'):

    # Load DataConfig
    data_config = rtc.DataConfig(input_folder)

    # Load the raw timeseries_import.xml
    raw_timeseries_import = pi.Timeseries(data_config, input_folder,
        basename_import, binary=False, pi_validate_times=True)

    # Make a fresh timeseries import that is a copy of the raw one
    new_filename = basename_import + '-polder_scheduling'
    timeseries_import = _duplicate_timeseries_import(
        raw_timeseries_import, new_filename, data_config, input_folder)


    # Set up some useful variables
    polder_outflows = []
    tide = timeseries_import.get(tide_timeseries)
    try:
        assert np.all(np.isfinite(tide)) # rudimentary data validation
    except AssertionError:
        logger.error('Gouda Polder Scheduling AssertionError: Eindk_Gouda_H: timeseries not complete.')
        logger.error('Eindk_Gouda_H = {}'.format(tide))
        raise
    last_index = len(tide) - 1
    dt = timeseries_import.dt.total_seconds()
    one_hour_integer_offset = int(3600.0 / dt)
    # make sure offset is always at least one
    one_hour_integer_offset = one_hour_integer_offset if one_hour_integer_offset > 0 else 1

    # Get indices of peak tide
    try:
        [peak_tide_indices] = np.where(np.isfinite(timeseries_import.get(tide_peak_timeseries)))
        [low_tide_indices] = np.where(np.isfinite(timeseries_import.get(tide_low_timeseries)))
    except IndexError:
        logger.error('Polder Scheduler IndexError: {}, {}: timeseries not complete.'.format(
                    tide_peak_timeseries, tide_low_timeseries))
        raise
    except KeyError:
        # Need to do signal processing to determine peak tide
        # Note: This can introduce a point of failure in the logic- use only as a last option
        #       We assume the tidal signal is strong and not noisy
        logger.warning('Polder Scheduler KeyError: {}, {}: timeseries not complete.'.format(
                    tide_peak_timeseries, tide_low_timeseries))
        [peak_tide_indices] = scipy.signal.argrelmax(tide, order=4)
        [low_tide_indices] = scipy.signal.argrelmin(tide, order=4)
    try:
        if low_tide_indices[0] != 0 and peak_tide_indices[0] != 0:
            # first step is unknown- need to decide what the first step is
            if low_tide_indices[0] < peak_tide_indices[0]:
                # for the purposes of this run, treat first step like a high tide
                peak_tide_indices = np.insert(peak_tide_indices, 0, 0)
            else:
                # for the purposes of this run, treat first step like a low tide
                low_tide_indices = np.insert(low_tide_indices, 0, 0)

        if low_tide_indices[-1] != last_index and peak_tide_indices[-1] != last_index:
            # last step is unknown- need to decide what the last step is
            if low_tide_indices[-1] > peak_tide_indices[-1]:
                # for the purposes of this run, treat last step like a high tide
                peak_tide_indices = np.append(peak_tide_indices, last_index)
            else:
                # for the purposes of this run, treat last step like a low tide
                low_tide_indices = np.append(low_tide_indices, last_index)

        # Find times between low and high tides
        # Note: We include an extra hour beyond the actual peak tide
        if low_tide_indices[0] == 0:
            rising_tide_indices = np.concatenate(
                [range(l, h + one_hour_integer_offset) for l, h in zip(low_tide_indices, peak_tide_indices)])
        else:
            rising_tide_indices = np.concatenate(
                [range(l, h + one_hour_integer_offset) for l, h in zip(low_tide_indices, peak_tide_indices[1:])])
    except:
            logger.error('Polder Scheduler IndexError: {}, {}: timeseries not complete.'.format(
                    tide_peak_timeseries, tide_low_timeseries))
            raise
            
    # Determine indices when polder pumping is allowed.
    polder_pumping_not_allowed_indices = np.union1d(np.where(tide < 0.0)[0], rising_tide_indices)
    polder_pumping_not_allowed_indices = polder_pumping_not_allowed_indices[polder_pumping_not_allowed_indices <= last_index]

    # Don't pump for one hour preceding the polder_pumping_not_allowed_indices, so that we can flush for that hour
    disallowed_indices = []
    for i in range(one_hour_integer_offset + 1):
        disallowed_indices.append(polder_pumping_not_allowed_indices - i)
    polder_pumping_not_allowed_indices =  np.unique(np.array(disallowed_indices, dtype=np.int))
    # drop any invalid new indices
    polder_pumping_not_allowed_indices = polder_pumping_not_allowed_indices[0 <= polder_pumping_not_allowed_indices]

    # Indices to divide up inflow
    inflow_seperation_indices = peak_tide_indices
    if inflow_seperation_indices[0] == 0:
        inflow_seperation_indices = inflow_seperation_indices[1:]
    if inflow_seperation_indices[-1] == last_index:
        inflow_seperation_indices = inflow_seperation_indices[:-1]

    # Allocate polder inflow into polder outflow
    for polder in polders:
        try:
            inflow = timeseries_import.get('BUF_{}_Q'.format(polder))
        except KeyError:
            inflow = np.full_like(tide, 0.0) # some default value
        try:
            max_outflow = timeseries_import.get('PG_{}_Qmax'.format(polder))
        except KeyError:
            max_outflow = np.full_like(tide, 1.2345) # some default value
        try:
            initial_outflow = timeseries_import.get('PG_{}_Q0_guess'.format(polder))[0]
        except KeyError:
            initial_outflow = 0.0 # some default value

        if initial_outflow == 0.0 and peak_tide_indices[0] == 0:
            # The actual peak tide is at the first time step, so we shift releases forward
            # Warning: This might be overly conservative if dt < 3600 s
            outflow_initial_placement_indices = np.insert(inflow_seperation_indices, 0, one_hour_integer_offset)
        else:
            # The first peak tide is not the real peak tide- no need to delay the polder release
            outflow_initial_placement_indices = np.insert(inflow_seperation_indices, 0, 0)

        # Set max to zero where pumping not allowed
        # Note: Here we are modifying the raw timeseries- Should we make a copy?
        max_outflow[polder_pumping_not_allowed_indices] = 0.0

        split_inflow = np.array_split(inflow, inflow_seperation_indices)
        sum_split_inflow = np.array([np.sum(s) for s in split_inflow])

        # Place total outflow per period at the peak tides
        outflow = np.zeros_like(inflow)
        outflow[outflow_initial_placement_indices] = sum_split_inflow

        # Make initial outflow match the provided initial outflow
        max_outflow[0] = initial_outflow

        if initial_outflow > 0.0 and polder_pumping_not_allowed_indices[0] == 0:
            logger.warning('Polder Scheduler: {} illegally released {} m3 at first timestep.'.format(
                    polder, dt * initial_outflow))
            outflow[outflow_initial_placement_indices[1]] -= initial_outflow
            outflow[0] = initial_outflow
        if initial_outflow > outflow[0]:
            # Polder initially released too much for this tidal cycle- reduce the volume at the next release cycle
            logger.warning('Polder Scheduler: {} released {} m3 too much at first timestep.'.format(
                    polder, dt * (initial_outflow - outflow[0])))
            outflow[outflow_initial_placement_indices[1]] -= initial_outflow - outflow[0]
            outflow[0] = initial_outflow

        # Iterate over outflow series, shifting outflow forward when it exceeds max_polder_outflow
        # Note: This will carry over any excess flow into next tidal period
        for i in range(last_index): # Note: this iterable does not include last index
            if outflow[i] > max_outflow[i]:
                outflow[i + 1] += outflow[i] - max_outflow[i]
                outflow[i] = max_outflow[i]

        # Drop any extra outflow
        if outflow[-1] > max_outflow[i]:
            logger.warning('Polder Scheduler: {} had {} m3 unreleased volume at last timestep.'.format(
                polder, dt * (outflow[-1] - max_outflow[-1])))
            outflow[-1] = max_outflow[i]

        # Store Result
        timeseries_import.set('PG_{}_Q'.format(polder), outflow)
        polder_outflows.append(outflow)

    # Now Determine Flushing Requirement for GEM
    polder_net_flow = np.sum(np.array(polder_outflows), axis=0)
    [polder_is_flowing] = np.where(np.logical_or(*polder_outflows))

    need_flushing_indices = []
    # Expand Backwards two hours backwards (so that start time is one hour before high tide)
    for i in range(one_hour_integer_offset * 2 + 1):
        need_flushing_indices.append(polder_is_flowing - i)
    # Expand forwards by one hour (must run one hour after last polder)
    for i in range(one_hour_integer_offset + 1):
        need_flushing_indices.append(polder_is_flowing + i)

    need_flushing_indices = np.unique(np.array(need_flushing_indices))
    # drop any invalid new indices
    need_flushing_indices = need_flushing_indices[0 <= need_flushing_indices]
    need_flushing_indices = need_flushing_indices[need_flushing_indices <= last_index]

    # Create a minimum flushing request timeseries
    # Nan indicate no minimum
    flushing_req = np.full_like(inflow, np.nan, dtype=np.float)
    flushing_req[need_flushing_indices] = 25.0
    try:
        previous_flushing_req = timeseries_import.get(flushing_pump_timseries_req)
    except KeyError:
        previous_flushing_req = np.full_like(inflow, np.nan, dtype=np.float)
    with np.warnings.catch_warnings():
        np.warnings.filterwarnings('ignore', r'All-NaN (slice|axis) encountered')
        new_flushing_req = np.nanmax([previous_flushing_req, flushing_req], axis=0)
    timeseries_import.set(flushing_pump_timseries_req,  new_flushing_req)

    timeseries_import.write()

    # Write out diff-able (sorted) original ts for debugging
    # new_filename = basename_import + '2'
    # timeseries_import2 = _duplicate_timeseries_import(
    #     raw_timeseries_import, new_filename, data_config, input_folder)
    # timeseries_import2.write()

def _duplicate_timeseries_import(old_timeseries_import, new_filename, data_config, input_folder):
    """
    Make a fresh timeseries import exactly like the old one
    """

    # Make an empty modified_timeseries_import
    new_timeseries_import = pi.Timeseries(data_config, input_folder,
        new_filename, binary=False, make_new_file=True)

    # Copy metadata over
    new_timeseries_import.times = \
        old_timeseries_import.times[old_timeseries_import.forecast_index:]
    new_timeseries_import.forecast_datetime = old_timeseries_import.forecast_datetime
    new_timeseries_import.dt = old_timeseries_import.dt
    new_timeseries_import.timezone = old_timeseries_import.timezone
    new_timeseries_import.ensemble_size = old_timeseries_import.ensemble_size
    new_timeseries_import.contains_ensemble = old_timeseries_import.contains_ensemble

    # Copy timeseries over
    for ts_name, values in old_timeseries_import.items():
        try:
            data_config.pi_variable_ids(ts_name)
        except KeyError:
            logger.warning('polder_scheduling.py: variable {} has no mapping defined in '
                'rtcDataConfig and cannot be added to {}.xml.'.format(ts_name, new_filename))
            continue
        else:
            new_timeseries_import.set(ts_name, values, unit=old_timeseries_import.get_unit(ts_name))

    return new_timeseries_import

if __name__ == '__main__':
    from Rijnland_Boezem_Optimization import Rijnland_base
    schedule_polder_pumps('../input', Rijnland_base.polders)