import csv
import logging
import os

import dill

import numpy as np

from rtctools.optimization.homotopy_mixin import HomotopyMixin
from rtctools.util import run_optimization_problem

# Useful extra funtions written by Deltares for Rijnland, kept in seperate file for readability
# Note: utilities also defines global vars BASE_FOLDER, MODEL_FOLDER, INPUT_FOLDER, OUTPUT_FOLDER, logger
from utilities import *


class PumpingAnalysis:

    def __init__(self, **kwargs):
        # Call parent class first for default behaviour.
        super().__init__(**kwargs)

        # Initialize class with results of previous run
        self._GoalProgrammingMixin__results = [self.previous_prio_results]
        self._GoalProgrammingMixin__results_are_current = True
        self._HomotopyMixin__theta = 1.0

    def goals(self):
        # Only use goals that are part of the analysis
        # g = super().goals()
        pump_goals = self.pump_cost_goals

        return pump_goals

    def path_goals(self):
        # Only use goals that are part of the analysis
        # g = super().path_goals()
        pump_goals = self.pump_cost_path_goals

        return pump_goals

    def priority_started(self, priority):
        # Make the seed method seed from superproblem results
        self._GoalProgrammingMixin__first_run = False

    def priority_completed(self, priority):
        pass

    def constraints(self, ensemble_member):
        constraints = super().constraints(ensemble_member)
        for func, func_min, func_max in self.superproblem_constraints:
            path_constraints.append(
                    (dill.loads(func)(self), func_min, func_max)
            )
        return constraints

    def path_constraints(self, ensemble_member):
        path_constraints = super().path_constraints(ensemble_member)
        for func, func_min, func_max in self.superproblem_path_constraints:
            path_constraints.append(
                    (dill.loads(func)(self), func_min, func_max)
            )
        return path_constraints

    def seed(self, ensemble_member):
        seed = super(HomotopyMixin, self).seed(ensemble_member)
        return seed

    def optimize(
        self, preprocessing=True, postprocessing=True, log_solver_failure_as_error=True
    ):
        # TODO: if HomotopyMixin is not a parent, just call super().
        # Note: This improvement is not relevant for Rijnland.
        success = super(HomotopyMixin, self).optimize(
            preprocessing=preprocessing,
            postprocessing=postprocessing,
            log_solver_failure_as_error=log_solver_failure_as_error,
        )
        return success

    def post(self):
        # Determine if final run was successful
        try:
            successfull_return_strings = {"Solve_Succeeded", "Solved_To_Acceptable_Level"}
            success = self.solver_stats["return_status"] in successfull_return_strings
        except AttributeError as error:
            success = False
            logger.warning('solver_stats attribute error = {}'.format(str(error)))

        if not success:
            logger.warning("What-if analysis failed at priority {}. Some solution path results will be missing.".format(self.this_priority))
        else:
            logger.info('success = {}'.format(str(success)))
            pump_costs = []
            total_time = self.times()[-1] - self.times()[0]
            for pump in sorted(self.pumps):
                try:
                    normalized_cost = self.get_timeseries(pump + "_minimized_cost").values[0]
                except KeyError:
                    pump_costs.append(0.0)
                    continue
                # Store costs, multiplying by duration to get true cost
                pump_costs.append(normalized_cost * total_time)

            # Aggregate GEM_Gouda into one pump
            gem_gouda_idx = sorted(self.pumps).index("GEM_Gouda")
            gem_gouda_inflow_idx = sorted(self.pumps).index("GEM_Gouda_inflow")
            pump_costs[gem_gouda_idx] += pump_costs[gem_gouda_inflow_idx]
            pump_costs[gem_gouda_inflow_idx] = 0.0

            with open(
                os.path.join(OUTPUT_FOLDER, "pump_costs.csv"), "a+"
            ) as csv_file:
                writer = csv.writer(csv_file)
                writer.writerow([self.this_priority] + pump_costs)



def execute_pumping_analysis(attribs):
    from Rijnland_Boezem_Optimization import Rijnland_Optimization

    # Run pump cost analysis
    opt_problem = type("PumpingAnalysis", (PumpingAnalysis, Rijnland_Optimization), attribs)
    run_optimization_problem(opt_problem, log_level=logging.INFO)
