"""
Useful extra funtions written by Deltares for Rijnland, kept in seperate file for readability
"""

import csv
import os
import logging
import sys
import numpy as np
import rtctools.data.pi as pi
from rtctools.optimization.timeseries import Timeseries


# Set up logging
logger = logging.getLogger("rtctools")


# Some handy folder vars
BASE_FOLDER = os.path.realpath(os.path.join(sys.path[0], '..'))
MODEL_FOLDER = os.path.join(BASE_FOLDER, 'model')
INPUT_FOLDER = os.path.join(BASE_FOLDER, 'input')
OUTPUT_FOLDER = os.path.join(BASE_FOLDER, 'output')

# Add stream handler if it does not already exist.
if not any((isinstance(h, logging.StreamHandler) for h in logger.handlers)):
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)

# Add pi.DiagHandler, but only if it does not already exist.
if not any((isinstance(h, pi.DiagHandler) for h in logger.handlers)):
    handler = pi.DiagHandler(OUTPUT_FOLDER)
    logger.addHandler(handler)

# Set initial log level to info
logger.setLevel(logging.INFO)


def snap_to_discrete(values, discrete_setpoints, strategy='closest', relaxaton_tolerance=0.5, pidt=None):
    """
    A function to snap pump bounds to discrete levels
    """
    if isinstance(values, Timeseries):
        return Timeseries(values.times, snap_to_discrete(values.values, discrete_setpoints, strategy, relaxaton_tolerance))
    else:
        values = np.array(values)
        setpoints = np.sort(np.array(discrete_setpoints))

        if strategy in ['up', 'down']:
            # we need to either snap up or down, but we also need to allow a
            # certain tol so that the snap is a relaxation if it is sufficiently
            # close to a bound

            # ensure tol is not larger than half the distance between the two closest setpoints
            tol = min(np.min(np.abs(np.diff(setpoints)))/2.0, abs(relaxaton_tolerance))

            # all points on and outside bounds are snapped to upper and lower bounds
            values[values >= (setpoints[-1] - tol)] = setpoints[-1]
            values[values <= (setpoints[0]  + tol)] = setpoints[0]

            # snap interior points in the appropriate direction
            if strategy == 'up':
                # Snap the values upwards, subject to relaxation_tolerance
                for sp_d, sp_u in zip(setpoints[:-1], setpoints[1:]):
                    values[(values >= sp_d + tol) & (values < sp_u + tol)] = sp_u
            elif strategy == 'down':
                # Snap the values downwards, subject to relaxation_tolerance
                for sp_d, sp_u in zip(setpoints[:-1], setpoints[1:]):
                    values[(values > sp_d - tol) & (values <= sp_u - tol)] = sp_d
            return values

        elif strategy == 'closest':
            for idx, value in enumerate(values):
                if np.isnan(value):
                    continue
                values[idx] = setpoints[np.abs(setpoints-value).argmin()]
            return values

        elif strategy == 'pid':
            raise NotImplementedError
        else:
            raise ValueError('Unknown snap strategy {}'.format(strategy))

# PID-like controller (to take into account accumulated error)
# def discrete_setpoint_controller(optimal_pumping, stages, p_gain = 1.0, i_gain = 0.1):
def discrete_setpoint_controller(optimal_pumping, stages,
                                 p_gain = 0.0, i_gain = 1.0/3600.0, d_gain = 0.0, dt = 1.0,
                                 min_flow=None, max_flow=None):
    i_error = 0.0
    previous_error = 0.0

    # Initialize outputs as empty arrays
    discrete_setpoints = np.empty_like(optimal_pumping, dtype=np.float)
    optimal_modes = np.empty_like(optimal_pumping, dtype=np.float)

    # Ensure that min and max arrays are initialized
    if min_flow is None:
        min_flow = np.full_like(optimal_pumping, -np.inf)
    if max_flow is None:
        max_flow = np.full_like(optimal_pumping, np.inf)

    for idx, target in enumerate(optimal_pumping):
        # Note: Min and max become feasible stage when within the stage min and max
        overshoot  = np.min(np.append(stages[stages[:, 1] >= target, 1],
                                      np.nanmin([np.max(stages[:, 1]), max_flow[idx]])) )
        undershoot = np.max(np.append(stages[stages[:, 1] <= target, 1],
                                      np.nanmax([np.min(stages[:, 1]), min_flow[idx]])) )

        overshoot_error =  abs(  (overshoot - target) * p_gain +
                                 (i_error + (overshoot - target)*dt) * i_gain +
                                 ((overshoot - target) - previous_error) / dt * d_gain  )
        undershoot_error = abs(  (undershoot - target) * p_gain +
                                 (i_error + (undershoot - target)*dt) * i_gain +
                                 ((undershoot - target) - previous_error) / dt * d_gain  )

        if overshoot_error > undershoot_error:
            setpoint = undershoot
        else:
            setpoint = overshoot
        discrete_setpoints[idx] = setpoint
        # If stage is not found, return miss value
        mode = stages[np.where(stages[:, 1] == setpoint), 0].flatten()
        optimal_modes[idx] = -999 if len(mode) != 1 else float(mode)
        error = (setpoint - target)
        i_error += error * dt
        previous_error = error
    return optimal_modes, discrete_setpoints

def process_sol_path_output(goal_list, solution_path_dict, file_path, write_mode='w'):
    """
    This is a function designed to take the solution path dict and a list of
    the goals that were used and generate a sol_path csv in the format
    requested by Ivo.
    """

    # Python does not have switches, so we simulate a switch with an if-else
    # block inside a function. Returns a list with values corresponding to the
    # items in entry_header
    entry_headers = ['element', 'doel', 'locatie', 'waarde', 'info']
    def get_goal_summary(goal, this_prio_sol, general=False):
        # Get the name of the goal
        goal_name = goal.__class__.__name__
        goal_name_Dutch = { 'WaterLevelRangeGoal'                   :'GrenzenRBP',
                            'WaterLevelSetpointGoal'                :'StreefpeilRBP',
                            'PumpDownstreamWaterLevelRangeGoal'     :'PompMinderEnAfslagPeilen',
                            'ChlorideRangeGoal'                     :'ChlorideMaximum',
                            'MinimizeQpumpGoal'                     :'MinimaliseerPompinzet',
                            'MinimizePumpCost'                      :'MinimaliseerPompkosten',
                            'QP_ROC_Goal'                           :'BeperkVariatiePompinzet',
                            'WeighPumps'                            :'WeegPompenAf',
                            'DischargeGoal'                         :'Afvoer'}

        # Get the name of the base element the goal variable is targeting,
        # which carries location info
        try:
            element_basename = goal.state.split('.')[0]
        except AttributeError:
            element_basename = goal.pump_var.replace('_Q', '')

        try:
            # There are a lot of lookups in this 'switch' that can result in
            # KeyErrors. To be a bit more graceful, we catch all KeyErrors
            if general:
                return ['Algemene informatie',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        'N/A',
                        'Doel is toegevoegd aan het optimalisatieprobleem']

            elif goal_name in 'WaterLevelRangeGoal':
                return ['Deel periode gehaald',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        round(this_prio_sol['proportion_level_satisfied_RBP']),
                        'Deel van de tijd dat het doel gehaald is']

            elif goal_name in 'WaterLevelSetpointGoal':
                return ['Deel periode gehaald',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        round(this_prio_sol['proportion_sp_satisfied_RBP']),
                        'Deel van de tijd dat het doel gehaald is']

            elif goal_name in 'PumpDownstreamWaterLevelRangeGoal':
                return ['Deel periode gehaald',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        round(this_prio_sol[element_basename + '_Hdown_min_proportion_satisfied']),
                        'Deel van de tijd dat het doel gehaald is']

            elif goal_name in 'ChlorideRangeGoal':
                return ['Deel periode gehaald',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        this_prio_sol[element_basename + '_Cl_prop_satisfied'],
                        'Deel van de tijd dat het doel gehaald is']

            elif goal_name in 'MinimizeQpumpGoal':
                return ['Uitgepompt volume (m^3)',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        round(this_prio_sol[element_basename + '_Qnet_out']),
                        'Totaal volume uitgepompt door dit gemaal']

            elif goal_name in 'MinimizeQpumpGoal':
                return ['Ingelaten volume (m^3)',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        round(this_prio_sol[element_basename + '_Qnet_in']),
                        'Totaal volume ingelaten door dit gemaal']

            elif goal_name in 'MinimizePumpCost':
                return ['Totale pompkosten (euro)',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        round(this_prio_sol[element_basename + '_net_cost_out']),
                        'Totale kosten gebruik van dit gemaal']

            elif goal_name in 'DischargeGoal':
                return ['RMS Error (m^3/s)',
                        goal_name_Dutch[goal_name],
                        element_basename,
                        round(this_prio_sol[element_basename + '_Qreq_RMS_error']),
                        'RMS error of the flow rate relative to the request. TODO: Dutch']


            else:
                logger.info(goal_name + ' not implemented in solution path summary.')
                return None

        except KeyError as e:
            logger.info('KeyError while generating sol_path_summary.csv: ' + str(e))
            return None

    # Open file
    with open(file_path + '/sol_path_summary.csv', write_mode) as f:
        # make csv writer
        writer = csv.writer(f)

        # write headers
        writer.writerow(['theta', 'prioriteit'] + entry_headers)

        # Iterate over solution path nested dictionary
        for theta, theta_results in sorted(solution_path_dict.items(), key = lambda x: str(x)):
            for priority, priority_results in sorted(theta_results.items(), key = lambda x: str(x)):

                # Get a list of the goals at this priority
                prio_goals = sorted([g for g in goal_list if int(g.priority) == priority], key = lambda x: str(x))

                # Add to the csv file for each goal
                for g in prio_goals:
                    general_entry = get_goal_summary(g, priority_results, general=True)
                    entry = get_goal_summary(g, priority_results)
                    if entry is None:
                        continue
                    writer.writerow([theta, priority] + general_entry)
                    writer.writerow([theta, priority] + entry)

                # Add Intermediate results
                prefix = [theta, priority, 'Tussentijdse resultaten']

                rbp_proportion_satisfied = [
                    'Waterstandsgrenzen',
                    'RBP',
                    round(priority_results['proportion_level_satisfied_RBP'], 3),
                    'Deel van de tijd dat het doel gehaald is']
                writer.writerow(prefix + rbp_proportion_satisfied)

                rbp_sp_proportion_satisfied = [
                    'Streefpeil',
                    'RBP',
                    round(priority_results['proportion_sp_satisfied_RBP'], 3),
                    'Deel van de tijd dat het doel gehaald is']
                writer.writerow(prefix + rbp_sp_proportion_satisfied)

                overall_pump_cost_out = [
                    'MinimaliseerPompkosten',
                    'alle gemalen (uitlaat)',
                    round(np.nansum([v for k, v in priority_results.items() if '_net_cost_out' in k])),
                    'Totale kosten (euro) gebruik van alle uitlaat pompen']
                writer.writerow(prefix + overall_pump_cost_out)

                overall_pump_flow_out = [
                    'MinimaliseerPompkosten', # This is the actual goal that minimizes the flow
                    'alle gemalen',
                    round(sum([v for k, v in priority_results.items() if '_Qnet_out' in k])),
                    'Totaal volume (m^3) uitgepompt door alle pompen']
                writer.writerow(prefix + overall_pump_flow_out)

                overall_pump_flow_in = [
                    'MinimaliseerPompkosten', # This is the actual goal that minimizes the flow
                    'alle gemalen',
                    round(sum([v for k, v in priority_results.items() if '_Qnet_in' in k])),
                    'Totaal volume (m^3) ingelaten door alle pompen']
                writer.writerow(prefix + overall_pump_flow_in)

                for key in [k for k in priority_results.keys() if '_Qnet_out' in k]:
                    location = key.replace('_Qnet_out', '')
                    q_net_out = [
                        'MinimaliseerPompkosten',
                        location,
                        round(priority_results[key]),
                        'Volume (m^3) uitgepompt bij ' + location]
                    writer.writerow(prefix + q_net_out)

                for key in [k for k in priority_results.keys() if '_Qnet_in' in k]:
                    location = key.replace('_Qnet_in', '')
                    q_net_in = [
                        'MinimaliseerPompkosten',
                        location,
                        round(priority_results[key]),
                        'Volume (m^3) ingelaten bij ' + location]
                    writer.writerow(prefix + q_net_in)

                for key in [k for k in priority_results.keys() if '_Cl_max' in k]:
                    location = key.replace('_Cl_max', '')
                    cl_max = [
                        'ChlorideMaximum',
                        location,
                        round(priority_results[key] * 1000),
                        'Maximale Cl concentratie (mg/L) bij ' + location]
                    writer.writerow(prefix + cl_max)

                for key in [k for k in priority_results.keys() if '_Cl_mean' in k]:
                    location = key.replace('_Cl_mean', '')
                    cl_mean = [
                        'ChlorideGemiddelde',
                        location,
                        round(priority_results[key] * 1000),
                        'Gemiddelde Cl concentratie (mg/L) bij ' + location]
                    writer.writerow(prefix + cl_mean)

                for key in [k for k in priority_results.keys() if '_Cl_prop_satisfied' in k]:
                    location = key.replace('_Cl_prop_satisfied', '')
                    cl_prop_satisfied = [
                        'ChlorideBinnenGrenzen',
                        location,
                        round(priority_results[key]),
                        'Deel Cl concentratie (mg/L) binnen grenzen bij ' + location]
                    writer.writerow(prefix + cl_prop_satisfied)

                for key in [k for k in priority_results.keys() if '_net_cost_out' in k]:
                    location = key.replace('_net_cost_out', '')
                    net_cost_out = [
                        'MinimaliseerPompkosten',
                        location,
                        round(priority_results[key]),
                        'Kosten (euro) voor gebruik uitlaat pompen ' + location]
                    writer.writerow(prefix + net_cost_out)

                for key in [k for k in priority_results.keys() if '_Qmean_out' in k]:
                    location = key.replace('_Qmean_out', '')
                    Qmean_out = [
                        '',
                        location,
                        round(priority_results[key]),
                        'Gemiddelde uitlaat (m^3/s) bij ' + location]
                    writer.writerow(prefix + Qmean_out)

                for key in [k for k in priority_results.keys() if '_Qreq_RMS_error' in k]:
                    location = key.replace('_Qreq_RMS_error', '')
                    Qmean_out = [
                        'Afvoer',
                        location,
                        round(priority_results[key], 2),
                        'Afvoer RMS Error (m^3/s) bij ' + location]
                    writer.writerow(prefix + Qmean_out)

                iter_count = [
                    'Iteraties',
                    '',
                    priority_results['iter_count'],
                    'Iteraties tot convergentie']
                writer.writerow(prefix + iter_count)

                wall_time = [
                    'Draaitijd',
                    '',
                    round(priority_results['wall_time'], 3),
                    'Totale tijd tot convergentie']
                writer.writerow(prefix + wall_time)

def sol_path_entry(opt_prob, results, custom_times = None, solver_stats = {}):

    if custom_times is None:
        times = opt_prob.times
    else:
        def times(*args, **kwargs):
            return custom_times

    # Get representative water level
    RBP_H = results['RBP_H']

    # Save water level range satisfaction
    satisfied_steps = np.less_equal(opt_prob._H_min, RBP_H) & np.greater_equal(opt_prob._H_max, RBP_H)
    n_sp_satisfied_RBP = np.sum(satisfied_steps)
    # TODO: Check if this works with variable timesteps
    proportion_level_satisfied_RBP = np.sum(satisfied_steps[1:] * np.diff(times())) / times()[-1]

    # Save water level setpoint satisfaction
    satisfied_steps = np.less_equal(opt_prob._H_sp_min, RBP_H) & np.greater_equal(opt_prob._H_sp_max, RBP_H)
    n_level_satisfied_RBP = np.sum(satisfied_steps)
    # TODO: Check if this works with variable timesteps
    proportion_sp_satisfied_RBP = np.sum(satisfied_steps[1:] * np.diff(times())) / times()[-1]

    entry = dict(
        process_time = solver_stats.get('t_mainloop.proc', -999.0),
        wall_time = solver_stats.get('t_mainloop.wall', -999.0),
        iter_count = solver_stats.get('iter_count', 'N/A'),
        n_level_satisfied_RBP = n_level_satisfied_RBP,
        proportion_level_satisfied_RBP = proportion_level_satisfied_RBP,
        proportion_sp_satisfied_RBP = proportion_sp_satisfied_RBP,
        n_sp_satisfied_RBP = n_sp_satisfied_RBP
        )

    # Store Cl results
    for cn in opt_prob.storage_nodes:
        try:
            cn_target_max = opt_prob.get_timeseries(cn + '_Cmax').values
        except KeyError:
            continue
        else:
            cn_result = results[cn + '.HQ.C']
            cn_satisfied_steps = np.bitwise_and(np.greater_equal(cn_result, 0.0), np.less_equal(cn_result, cn_target_max))
            entry[cn + '_Cl_n_satisfied'] = np.sum(cn_satisfied_steps)
            entry[cn + '_Cl_prop_satisfied'] = np.sum(cn_satisfied_steps[1:] * np.diff(times())) / times()[-1]
            entry[cn + '_Cl_mean'] = np.trapz(cn_result,  times()) / times()[-1]
            entry[cn + '_Cl_max'] = np.max(cn_result)

    for pump in opt_prob.pumps:
        if '_inflow' in pump:
            # Make sure that results were modified so that the internal
            # complementary pump is either zero or not in the results at all
            try:
                assert np.all(results[pump + '_Q'] == 0.0)
            except KeyError:
                pass
            # Skip the pump- it is not relevant to the sol path
            continue

        # TODO: Check if this works with variable timesteps and/or overridden pump timesteps

        # Add pump flows
        p_times = times(pump + '_Q')
        q_results = results[pump + '_Q']
        q_net = np.trapz(q_results, p_times)
        # Split negative (outwards) pump flows from positive (inwards) pump flows
        q_net_out = np.trapz(-np.minimum(q_results, 0.0), p_times)
        q_net_in = np.trapz(np.maximum(q_results, 0.0), p_times)
        entry[pump + '_Qnet_out'] = q_net_out
        entry[pump + '_Qnet_in'] = q_net_in
        entry[pump + '_Qmean_out'] = q_net_out / p_times[-1]
        entry[pump + '_Qmean_in'] = q_net_in / p_times[-1]
        entry[pump + '_Qmean'] = q_net / p_times[-1]

        # Process q request goals
        try:
            q_req = opt_prob.get_timeseries(pump + '_Qreq').values
        except KeyError:
            entry[pump + '_Qreq_n_timesteps'] = 0
            entry[pump + '_Qreq_n_sig_viol'] = 0.0
            entry[pump + '_Qreq_RMS_error'] = 0.0
            entry[pump + '_Qreq_max_error'] = 0.0
        else:
            # Number of states with a flow request goal
            n_req_steps = np.sum(np.isfinite(q_req))
            entry[pump + '_Qreq_n_timesteps'] = n_req_steps

            # Number of states with a "significant" violation of q_req_tol
            finite_error = np.nan_to_num(q_results - q_req)
            n_sig_viol = np.sum(np.fabs(finite_error) >= opt_prob.q_req_tol)
            entry[pump + '_Qreq_n_sig_viol'] = n_sig_viol

            # Root Mean Square Error of request goal
            rms_error = np.sqrt(np.sum(np.square(finite_error)) / n_req_steps)
            entry[pump + '_Qreq_RMS_error'] = rms_error

            # Max error
            entry[pump + '_Qreq_max_error'] = np.max(finite_error)

        # Add pump costs
        cost_series = results.get(pump + '_UnitPrice_m3s', np.nan)
        # Positive flows are passive and don't incur or reduce costs
        net_cost_out = np.trapz(-np.minimum(q_results, 0.0) * cost_series, p_times)
        entry[pump + '_net_cost_out'] = net_cost_out

        # Add results with regard to the PumpDownstreamWaterLevelRangeGoal
        try:
            lower_level_goal = opt_prob.get_timeseries(pump + '_Hdown_min').values - opt_prob.level_tol
            upper_level_goal = opt_prob.get_timeseries(pump + '_Hdown_max').values + opt_prob.level_tol
        except KeyError:
            pass
        else:
            h_up = results[pump + '.HQDown.H']
            entry[pump + '_Hdown_maximum_violation'] = min(0.0, np.max(h_up - lower_level_goal), np.max(upper_level_goal - h_up))
            satisfied_steps = np.less_equal(lower_level_goal, h_up) & np.greater_equal(upper_level_goal, h_up)
            n_level_satisfied = np.sum(satisfied_steps)
            # TODO: Check if this works with variable timesteps
            proportion_level_satisfied = np.sum(satisfied_steps[1:] * np.diff(times())) / times()[-1]
            entry[pump + '_Hdown_min_proportion_satisfied'] = proportion_level_satisfied

    return entry
